/* 
* Copyright (C) 2015, Adrian Jarabo (http://giga.cps.unizar.es/~ajarabo/)
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the "Software"),
* to deal in the Software without restriction, including without limitation
* the rights to use, copy, modify, merge, publish, distribute, sublicense,
* and/or sell copies of the Software, and to permit persons to whom
* the Software is furnished to do so, subject to the following conditions:

* The above copyright notice and this permission notice shall be included
* in all copies or substantial portions of the Software.

* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
* DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
* OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef __FILM_H
#define __FILM_H

//Deleting min and max macros (pain in the arse when using MSVC)
#ifdef min
#undef min
#endif
#ifdef max
#undef max
#endif

#include "Utils/globals.h"
#include "LinearAlgebra/Vector2.h"
#include "Filter/BoxFilter.h"
#include "Sampling/Sample.h"
#include "Integrator/RadianceSample.h"
#include "image/image.h"


#include <math.h>
#include <iostream>
#include <stdio.h>
#include <list>


/** A render Film for RGB images in Real format. */
class Film
{
public:
	enum Components { RADIANCE = 0, DEPTH = 1, VARIANCE = 2, ALL = RADIANCE | DEPTH | VARIANCE };

protected:
	/// The components stored by the film
	Components components;
	
	/// Reconstruction Filter
	Filter *filter;

	/// Actual Bitmap containing image
	Real *image;
	Real *weight;
	
	/// Buffer containing additional components
	Real *aux_comp;

	/// Width and height of Film
	int width, height;
	int shift_x, shift_y;
	Real ratio;

	/// Half width and height of Film.
	Real xw2,yw2;
	
	bool fixed;

	void get_film_samples(const Vector2 &position, std::list<Vector2> &samples)const
	{
		samples.clear();

		int size_kernel = filter->get_size();
		int half_size = size_kernel / 2;
		int x = static_cast<int>(position[0]);
		int y = static_cast<int>(position[1]);
		
		int init_x = x-half_size, end_x = x+half_size;
		int init_y = y-half_size, end_y = y+half_size;
		init_x = (init_x < 0 )?0:init_x;
		init_y = (init_y < 0 )?0:init_y;
		end_x = (end_x < width)?end_x:(width-1);
		end_y = (end_y < height)?end_y:(height-1);

		for( int j=init_y; j<=end_y; ++j )
		for( int i=init_x; i<=end_x; ++i )
			samples.push_back(Vector2(i+.5,j+.5));

	}
	void get_film_samples(const Real x, const unsigned int d, std::list<Real> &samples)const
	{
		samples.clear();

		int size_kernel = filter->get_size();
		int half_size = size_kernel / 2;
		int ix = static_cast<int>(x);
		
		int init_x = x-half_size, end_x = x+half_size;

		init_x = (init_x < 0 )?0:init_x;
		end_x = (end_x < (d?(width-1):(height-1)))?end_x:(d?(width-1):(height-1));

		for( int i=init_x; i<=end_x; ++i )
			samples.push_back(i+.5);
	}

	inline void draw_pixel(int x, int y, const Vector3& cv, Real w)
	{
		image[(y*width+x)*3] += cv[0];
		image[(y*width+x)*3+1] += cv[1];
		image[(y*width+x)*3+2] += cv[2];
		
		weight[(y*width+x)] += w;
	}
	void fix(Imaging::Image &out_image, Film::Components comp = Film::RADIANCE)
	{
		out_image = Imaging::Image(width,height);
		int nb_components = 0;
		Real *data = 0;

		// Determine the number of components to be stored in the RGB map...
		switch( comp )
		{
		// Store the RGB radiance
		case Film::RADIANCE:
			nb_components = 3;
			fixed = true; //se utiliz una extensin no estndar: enumeracin 'enumeracin' utilizada en el nombre completo
//nonstandard extension used: enum 'enum' used in qualified name
//Cuando se hace referencia a una enumeracin dentro de un tipo, no necesita especificar el nombre de la enumeracin.

			data = image;
			break;
		// Store float components
		case Film::DEPTH:
			nb_components = 1;
			data = aux_comp;
			break;
		// If not an individual component, ignore...
		default:
			return;
		}

		// Store all
		int index = 0, index_w = 0;
		int delta_c1 = 1%nb_components, delta_c2 = 2%nb_components;
		for(int i = (height - 1); i >= 0; --i)
		{
			for(int j = 0; j < width; ++j)
			{	
				if( weight[index_w] )
				{
					Real inv_weight = 1./weight[index_w];
					out_image.set(j,i, Imaging::RGBColor(data[index  ]*inv_weight,
														 data[index+delta_c1]*inv_weight,
														 data[index+delta_c2]*inv_weight));
				}
				index+=nb_components;
				++index_w; 
			}
		}
	}

public:
	/// Draw a pixel to the Film (deprecated...)
	inline void draw_pixel(int x, int y, const Vector3& cv)
	{
		image[(y*width+x)*3] += cv[0];
		image[(y*width+x)*3+1] += cv[1];
		image[(y*width+x)*3+2] += cv[2];
	}
	void init()
	{
		int idx_rgb = 0;
		for(int i=0; i<width*height;++i)
		{
			image[idx_rgb] = 0.;
			image[idx_rgb+1] = 0.;
			image[idx_rgb+2] = 0.;
			weight[i] = 0.;

			if( components )
				aux_comp[i] = 0.;

			idx_rgb += 3;
		}
	}
public:
	/// Constructs a Film with specified dimensions
	Film(int w, int h, Components comp = RADIANCE): 
		width(w), height(h), shift_x(0), shift_y(0), ratio(static_cast<Real>(width)/static_cast<Real>(height)), 
		image(0), weight(0), filter(new BoxFilter(1)), fixed(true),
		components(comp), aux_comp(0)
	{
		xw2 = width / 2.0f;
		yw2 = height / 2.0f;
		image = new Real[3*width*height]();
		weight = new Real[width*height]();

		if(components)
			aux_comp = new Real[width*height]();
		
		init();
	}
	
	Film(int w, int h, Filter *_filter, Components comp = RADIANCE): 
		width(w), height(h), shift_x(0), shift_y(0), 
		ratio(static_cast<Real>(width)/static_cast<Real>(height)), 
		image(0), weight(0), filter(_filter?_filter:(new BoxFilter(1))), fixed(true),
		components(comp), aux_comp(0)
	{
		xw2 = width / 2.0f;
		yw2 = height / 2.0f;
		image = new Real[3*width*height]();
		weight = new Real[width*height]();

		if(components)
			aux_comp = new Real[width*height]();

		init();
	}

	~Film()
	{
		clear();
	}
    virtual Real get_time_length() const {return 0.;}
    virtual Real get_time_offset() const {return 0.;}

	virtual int get_time_resolution() const {return 0;}

	/// Returns width of Film
	int get_width() const {return width;}

	/// Returns height of Film
	int get_height() const {return height;}
  
	/** 2D transform from window coordinates to image coordinates.
			Image coordinates are in the [-.5,.5]x[-.5/ratio,.5/ratio] range */
	const Vector2 window_coords2image_coords(const Vector2& p) const
	{
		return Vector2((p[0]-xw2+shift_x)/width, (p[1]-yw2+shift_y)/height / ratio);
	}
	
	void set_shift(const int sx, const int sy){shift_x=sx; shift_y=sy;}
	virtual void add_sample(const Sample &sample, const RadianceSample &radiance)
	{
		fixed = false;

		std::list<Vector2> film_samples;
		get_film_samples(sample.position,film_samples);
		//draw_pixel(x,y,radiance.radiance*sample.weight);

		std::list<Vector2>::iterator s;
		
		Vector3 rad = radiance.radiance.to_rgb();//*sample.weight;

		for( s = film_samples.begin(); s != film_samples.end(); s++ )
		{
			Real w = filter->evaluate((*s)-sample.position)*sample.weight;
			int sx = static_cast<int>((*s)[0]), sy = static_cast<int>((*s)[1]);
			draw_pixel(sx, sy, rad*w, w);
			if(components)
				aux_comp[(sy*width+sx)] += radiance.distance*w;
		}
	}

	virtual void add_samples(const Sample &sample, const std::list<RadianceSample> &radiances)
	{
	}
	void write(const char* name) 
	{
		char ncomp[180];
		strcpy(ncomp,name);
		
		// Store radiance image
		Imaging::Image out_image;
		fix(out_image);
		strcat(ncomp,".hdr");
		out_image.save(ncomp);

		// Store depth image
		if( components & DEPTH)
		{
			fix(out_image, DEPTH);
			strcpy(ncomp,name);
			strcat(ncomp,"_depth.hdr");
			out_image.save(ncomp);
		}

		//---------- OLD CODE FOR STORING TGA ----------------------------------------------
		//unsigned char TGAheader[12]= { 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
		//unsigned char header[6] = { width%256, width/256, height%256, height/256, 24, 0 };
		//unsigned char* pixels = new unsigned char[width * height * 3];
		//FILE *fScreenshot;
		//
		////put the color data into the pixels array
		//int index = 0;
		//int lineoffset = 0;
		//unsigned int pixel = 0;
		//unsigned int red = 0, green = 0, blue = 0;
		//unsigned int* ptr = (unsigned int *)image;
		//
		//for(int i = (height - 1); i >= 0; --i)
		//{
		//	for(int j = 0; j < width; ++j)
		//	{	
		//		unsigned char r = static_cast<unsigned char>(image[index] * 255.0f);
		//		unsigned char g = static_cast<unsigned char>(image[index+1] * 255.0f);
		//		unsigned char b = static_cast<unsigned char>(image[index+2] * 255.0f);

		//		pixels[(index) + 0] = (b>255)?255:b;
		//		pixels[(index) + 1] = (g>255)?255:g;
		//		pixels[(index) + 2] = (r>255)?255:r;
		//		
		//		index+=3;
		//	}
		//}
		//
		////write screenshot
		//fScreenshot = fopen(name, "wb");
		//fwrite(TGAheader, sizeof(unsigned char), 12, fScreenshot);
		//fwrite(header, sizeof(unsigned char), 6, fScreenshot);
		//fwrite(pixels, sizeof(unsigned char), (width * height * 3), fScreenshot);
		//fclose(fScreenshot);
		//
		//delete[] pixels;
	}

	virtual void clear()
	{
		if(image)
			delete[]image;

		if(weight)
			delete[]weight;

		if(aux_comp)
			delete[]aux_comp;
	}
	
	virtual Real get_exposure_time ()
	{ return 0.; }
};
#endif


