/*
* Copyright (C) 2015, Adrian Jarabo (http://giga.cps.unizar.es/~ajarabo/)
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the "Software"),
* to deal in the Software without restriction, including without limitation
* the rights to use, copy, modify, merge, publish, distribute, sublicense,
* and/or sell copies of the Software, and to permit persons to whom
* the Software is furnished to do so, subject to the following conditions:

* The above copyright notice and this permission notice shall be included
* in all copies or substantial portions of the Software.

* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
* DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
* OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef _KERNEL_BASED_STREAK_CAMERA_FILM_H_
#define _KERNEL_BASED_STREAK_CAMERA_FILM_H_

#include "Film/StreakCameraFilm.h"
#include "DataStructures/KDTree.h"
#include "DensityEstimation\DensityEstimationKernel.h"

#include <vector>

class KernelBasedStreakCameraFilm : public StreakCameraFilm
{
	std::vector<KDTree<Spectrum,1>*> m_available_data;
	std::vector<Real*> m_min_time_sample;

	unsigned int m_nb_nearest_photons;
	Real m_radius_scale, m_max_radius;
	bool m_include_direct;

	Spectrum get_radiance(const Real ts, const KDTree<Spectrum, 1> &kd_tree, const int nb_nearest_photons, const Real radius_scale, const Real max_radius)
	{
		Spectrum radiance(0);
		DensityEsimationKernels::Perlin<1>  Kt;

		if (kd_tree.is_empty())
			return Spectrum();

		if (nb_nearest_photons > 1)
		{
			// Compute initial radius...
			std::vector<const KDTree<Spectrum, 1>::Node*> photons;
			std::vector<const KDTree<Spectrum, 1>::Node*>::iterator ph;
			Real rad;

			std::vector<Real> v; v.push_back(ts);
			kd_tree.find(v, nb_nearest_photons, photons, rad);

			// Update radius
			rad *= radius_scale;

			if (max_radius < rad)
				rad = max_radius*radius_scale;

			int nb_before = 0;
			int nb_in_kernel = 0;

			// And compute contribution
			for (ph = photons.begin(); ph != photons.end(); ph++)
			{
				Real tp = (*ph)->point()[0];
				if (fabs(tp - ts) < rad)
				{
					++nb_in_kernel;
					radiance += ((*ph)->data())*Kt(fabs(tp - ts), rad);
				}
				if ((ts - tp) >= 0) ++nb_before;
			}
		}
		else
		{
			std::list<const KDTree<Spectrum, 1>::Node*> photons;
			std::list<const KDTree<Spectrum, 1>::Node*>::iterator ph;

			Real rad = max_radius*radius_scale;
			std::vector<Real> v; v.push_back(ts);
			kd_tree.find(v, rad, &photons);


			int nb_before = 0;
			int nb_in_kernel = 0;

			// And compute contribution
			for (ph = photons.begin(); ph != photons.end(); ph++)
			{
				Real tp = (*ph)->point()[0];
				++nb_in_kernel;
				radiance += ((*ph)->data())*Kt(fabs(tp - ts), rad);
				if ((ts - tp) >= 0) ++nb_before;
			}
		}

		return radiance;
	}
	
	Real compute_radius_scale(unsigned int it, Real alpha)
	{
		Real radius_scale = 1.;
		for (int i = 1; i <= it; ++i)
		{
			radius_scale *= (static_cast<Real>(i-1)+alpha) / static_cast<Real>(i);
		}
		return radius_scale;

	}
	void fix_and_store_slice(const int y)
	{
		Imaging::Image out_image(m_time_resolution, width);

		// First, fix the slice (which basicaly means that we are normalizing the 
		// values on it!
		Real *curr_slice = m_available_slices[y - m_first_available_slice];

		//int index_w = (y-m_first_available_slice)*width;
		int index_w = (y)*width;
		int index = 0;

		for (int i = 0; i<width; ++i)
		{
			m_available_data[y - m_first_available_slice][i].balance();
			if (weight[index_w])
			{
				Real inv_weight = 1. / weight[index_w];

				for (int t = 0; t < m_time_resolution; ++t)
				{
					Spectrum r = get_radiance(t, m_available_data[y - m_first_available_slice][i], m_nb_nearest_photons, m_radius_scale, m_max_radius);

					out_image.set(t, i, Imaging::RGBColor((r[0] + curr_slice[index]) * inv_weight,
						(r[1] + curr_slice[index+1]) * inv_weight,
						(r[2] + curr_slice[index+2]) * inv_weight));

					index += 3;
				}
				
			}
			else
				index += m_time_resolution * 3;

			++index_w;
		}

		//And after that, store it!
		char name[1024];
		char name_var[1024];
		char name_ns[1024];

		sprintf(name, "%s_%04d.%s", m_name.c_str(), y, m_extension.c_str());

		//workaround for hdr images less than 8 pixels wide (matlab fails reading them) => store them transposed 
		if (out_image.width() < 8) out_image.transpose();

		out_image.save(name);
	}
	
	void init()
	{
		if (m_time_resolution && width)
		{
			for (int i = 0; i < m_nb_available_slices; ++i)
			{
				m_available_slices.push_back(new Real[3 * m_time_resolution*width]());
				m_available_data.push_back(new KDTree<Spectrum, 1>[width]());
				m_min_time_sample.push_back(new Real[width]);
			}


		}
	}
public:
	KernelBasedStreakCameraFilm(int w, int h, int t, Real exposure, Filter *_filter, Real radius, Real alpha, unsigned int it, bool _camera_unwarp = false, Components comp = RADIANCE)
		:StreakCameraFilm(w, h, t, exposure, _filter, _camera_unwarp, comp), m_radius_scale(compute_radius_scale(it, alpha)), m_max_radius(radius), m_nb_nearest_photons(0), m_include_direct(false)
	{
		init();
	}

	KernelBasedStreakCameraFilm(int w, int h, int t, Real exposure, Filter *_filter, int nb_photons, Real alpha, unsigned int it, bool _camera_unwarp = false, Components comp = RADIANCE)
		:StreakCameraFilm(w, h, t, exposure, _filter, _camera_unwarp, comp), m_radius_scale(compute_radius_scale(it,alpha)), m_max_radius(1000), m_nb_nearest_photons(nb_photons), m_include_direct(false)
	{
		init();
	}

	~KernelBasedStreakCameraFilm()
	{
		clear();
	}

	virtual void add_samples(const Sample &sample, const std::list<RadianceSample> &radiances)
	{
		// First the film's samples are found...
		std::list<Vector2> film_samples;
		std::list<Real> temporal_samples;
		Film::get_film_samples(sample.position, film_samples);
		
		// We iterate over all film samples
		std::list<Vector2>::iterator s;
		std::list<Real>::iterator t;
		for (s = film_samples.begin(); s != film_samples.end(); s++)
		{
			int x = static_cast<int>((*s)[0]);
			int y = static_cast<int>((*s)[1]);

			//Check if new scanline is needed...
			if (y > m_last_available_slice)
			{
				//... if so, get it!
				//	First store the unused scanline...
				fix_and_store_slice(m_first_available_slice);

				//... and remove it from memory.
				delete[] m_available_slices.front();
				delete[] m_available_data.front();
				delete[] m_min_time_sample.front();
				m_available_slices.erase(m_available_slices.begin());
				m_available_data.erase(m_available_data.begin());
				m_min_time_sample.erase(m_min_time_sample.begin());

				// And then, add the new scanline
				m_available_slices.push_back(new Real[3 * m_time_resolution*width]());
				m_available_data.push_back(new KDTree<Spectrum, 1>[width]());
				m_min_time_sample.push_back(new Real[width]);

				++m_first_available_slice;
				++m_last_available_slice;
			}

			// We first get the weight of the pixel...
			//Filter in the y dimension...
			Real w = filter->evaluate((*s) - sample.position)*sample.weight;
			// ... and store it on weights matrix!
			weight[(y*width + x)] += w;



			// Now, we can iterate over all samples
			std::list<RadianceSample>::const_iterator r;

			for (r = radiances.begin(); r != radiances.end(); ++r)
			{
				// Get time_coordinate
				Real pos_sensor = ((camera_unwarp ? (r->time - r->distance) : r->time) - m_offset) / m_exposure_time;
				int idx = y - m_first_available_slice;

				if( m_min_time_sample[idx][x] > pos_sensor )
					m_min_time_sample[idx][x] = pos_sensor;

				if (r->bounce > 1 || m_include_direct)
				{
					std::vector<Real> v; v.push_back(pos_sensor);
					m_available_data[idx][x].store(v, r->radiance*w);
				}
				else
				{

					// If out of the range of the streak sensor, then discard.
					if (!(pos_sensor < m_time_resolution && pos_sensor >= 0)) continue;
					// Else, get the current slice, and the temporal samples in the streak film

					get_film_samples(pos_sensor, 2, temporal_samples);

					for (t = temporal_samples.begin(); t != temporal_samples.end(); ++t)
					{
						//m_available_slices[idx]->draw_pixel(x,static_cast<int>(*t),r->radiance*w*filter->get_subfilter(2)->evaluate(t-pos_sensor));
						Vector3 rad = r->radiance.to_rgb()*w*time_filter->evaluate((*t) - pos_sensor);
						int offset = (static_cast<int>(*t) + x*m_time_resolution) * 3;


						m_available_slices[idx][offset] += rad[0];
						m_available_slices[idx][offset + 1] += rad[1];
						m_available_slices[idx][offset + 2] += rad[2];

					}

				}
				//And store it into the accumulated image
				draw_pixel(x, y, r->radiance.to_rgb()*w);
			}
		}
	}


	virtual void clear()
	{
		if (!m_available_slices.empty())
		{
			for (int slice = m_first_available_slice; slice <= m_last_available_slice; ++slice)
				fix_and_store_slice(slice);

			m_last_available_slice = -1;

			while (!m_available_slices.empty())
			{
				if (m_available_slices.back())
					delete[] m_available_slices.back();

				m_available_slices.pop_back();
			}

			while (!m_available_data.empty())
			{
				if (m_available_data.back())
					delete[] m_available_data.back();

				m_available_data.pop_back();
			}
		}
	}
}; //StreakCameraFilm


#endif //_STREAK_CAMERA_FILM_H_