/* 
* Copyright (C) 2015, Adrian Jarabo (http://giga.cps.unizar.es/~ajarabo/)
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the "Software"),
* to deal in the Software without restriction, including without limitation
* the rights to use, copy, modify, merge, publish, distribute, sublicense,
* and/or sell copies of the Software, and to permit persons to whom
* the Software is furnished to do so, subject to the following conditions:

* The above copyright notice and this permission notice shall be included
* in all copies or substantial portions of the Software.

* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
* DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
* OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef _MULTIBOUNCE_STREAK_CAMERA_FILM_H_
#define _MULTIBOUNCE_STREAK_CAMERA_FILM_H_

#include "Film/StreakCameraFilm.h"
#include "Filter/UberFilter.h"
#include <vector>

class MultibounceStreakCameraFilm: public Film
{
	unsigned int m_nb_bounces;
	Real m_exposure_time;
	std::vector<StreakCameraFilm *> m_bounce_films;
	
	void init(int w, int h, int t, Real exposure, Filter *_filter, bool _camera_unwarp = false)
	{
		m_exposure_time = exposure;
		for(int b=0; b<m_nb_bounces; ++b)
		{
			m_bounce_films.push_back(new StreakCameraFilm(w,h,t,exposure,_filter, _camera_unwarp));
			
			char name[512];
			sprintf(name, "frame_b%02d", b+1);
			m_bounce_films.back()->set_name(name, "hdr");
		}
	}
public:
	MultibounceStreakCameraFilm(int w, int h, int t, Real exposure, unsigned int nb_bounces, 
								Components comp = RADIANCE)
		:Film(w, h, comp), m_nb_bounces(nb_bounces)
	{	
		init(w,h,t,exposure,this->filter);
	}
	MultibounceStreakCameraFilm(int w, int h, int t, Real exposure, unsigned int nb_bounces, 
								Filter *_filter, Components comp = RADIANCE)
		:Film(w, h,_filter, comp ), m_nb_bounces(nb_bounces)
	{	
		init(w,h,t,exposure,_filter);
	}
	
	MultibounceStreakCameraFilm(int w, int h, int t, Real exposure, unsigned int nb_bounces, 
								Filter *_filter, bool _camera_unwarp = false, Components comp = RADIANCE)
	:Film(w, h,_filter, comp ), m_nb_bounces(nb_bounces)
	{	
		init(w,h,t,exposure,_filter, _camera_unwarp);
	}

	~MultibounceStreakCameraFilm()
	{
		clear();
	}
	
    Real get_time_length() const {if(m_nb_bounces) return m_bounce_films.front()->get_time_length();  return 0;}
    Real get_time_offset() const {if(m_nb_bounces) return m_bounce_films.front()->get_time_offset();  return 0;}
    
	// Returns time resolution of StreakFilm
	int get_time_resolution() const {if(m_nb_bounces) return m_bounce_films.front()->get_time_resolution();  return 0;}
	//Allows setting the name of the streak images
	void set_name(const char *name, const char *ext)
	{
		//m_name = std::string(name); m_extension = std::string(ext);
		for(int b=0; b<m_nb_bounces; ++b)
		{	
			char nname[512];
			sprintf(nname, "%s_b%02d", name, b+1);
			m_bounce_films[b]->set_name(nname, ext);
		}
	}

	virtual void add_sample(const Sample &sample, const RadianceSample &radiance) //TODO
	{
	}
	virtual void add_samples(const Sample &sample, const std::list<RadianceSample> &radiances)
	{
		std::vector< std::list<RadianceSample> > bounce_radiances;
		bounce_radiances.resize(m_nb_bounces);

		std::list<RadianceSample>::const_iterator s;
		for( s = radiances.begin(); s != radiances.end(); s++ )
		{
			if( s->bounce < m_nb_bounces )
				bounce_radiances[s->bounce].push_back(*s);
			else
				bounce_radiances.back().push_back(*s);

			Film::add_sample(sample, *s);
		}

		for( int b=0; b<m_bounce_films.size(); ++b )
			m_bounce_films[b]->add_samples( sample, bounce_radiances[b] );
	}
	virtual void clear()
	{
		for(int b=0; b<m_nb_bounces; ++b)
			delete m_bounce_films[b];

		Film::clear();
	}
	virtual Real get_exposure_time ()
	{ return m_exposure_time; }
}; //MultibounceStreakCameraFilm


#endif //_MULTIBOUNCE_STREAK_CAMERA_FILM_H_