/* 
* Copyright (C) 2015, Adrian Jarabo (http://giga.cps.unizar.es/~ajarabo/)
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the "Software"),
* to deal in the Software without restriction, including without limitation
* the rights to use, copy, modify, merge, publish, distribute, sublicense,
* and/or sell copies of the Software, and to permit persons to whom
* the Software is furnished to do so, subject to the following conditions:

* The above copyright notice and this permission notice shall be included
* in all copies or substantial portions of the Software.

* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
* DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
* OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef _BIDIRECTIONAL_PATH_TRACING_INTEGRATOR_H_
#define _BIDIRECTIONAL_PATH_TRACING_INTEGRATOR_H_

#include "Integrator/Integrator.h"
#include "RayTracing/World.h"
#include "Sampling/SamplingTechniques.h"
#include "Path.h"
#include "PathMIS.h"

//=============================================================
template<int D>
class BidirectionalPathTracing: public Integrator<D>
{
public:
	enum NextEventSampling{ MeanFreePath, Time };
	enum AngularSampling{ PhaseFunction, ShortestTime };
	enum TerminationSampling{ RussianRoulette, NbBounces, TotalTime };
	enum VertexConnection{ ShadowConnection, TimeSampling };
	enum WeightingTechnique{ MIS_Balance, MIS_Power, Uniform };

private:
	NextEventSampling m_next_event_sampling_technique;
	AngularSampling m_angular_sampling_technique;
	TerminationSampling m_termination_sampling_technique;
	VertexConnection m_vertex_connection_technique;
	WeightingTechnique m_weighting_technique;
	Real m_beta_MIS;
	int m_avg_nb_vertices_path;
	int m_max_path_size;

	bool m_path_trace_only;
	bool m_keep_track_sample_count;
	// ------------------------------------------------------------------------------------------
    Film *film;
	
	unsigned int m_incoming_samples;
	Real m_inv_incoming_samples;
	FILE *f_log;
	
	// ------------------------------------------------------------------------------------------
	// Path Sampling Functions
	Real sample_next_event_distance(const Ray<D> &r, Real &p)const;
	void correct_next_event_distance( Ray<D> &r, Real &distance, bool &in_media, Real &p )const;
	void sample_outgoing_ray( Ray<D> &ray, Intersection<D> &it, bool media_interaction, 
							  Real sampled_distance, Real current_delay, const VectorN<D> &target, 
							  Spectrum &f, Real &time, Real &p )const;
	bool sample_termination( Real albedo, Real time, int nb_bounces, Real &p )const;
	void generate_path(const Ray<D> &r, const Spectrum &f0, const Real p0, const VectorN<D> &target, Path<D> &path )const;

	// ------------------------------------------------------------------------------------------
	// Path Connecting Functions
	bool connect_vertex_with_light( const typename Path<D>::Vertex *eye_vertex, 
									Spectrum &f, Real &p, Real *time = 0 )const;
	bool connect_vertices( const typename Path<D>::Vertex *eye_vertex, const typename Path<D>::Vertex *light_vertex, 
						   Spectrum &f, Real *time = 0, Real *pdf_time = 0 )const;
	Real compute_weights(const Path<D> &eye_path, const Path<D> &light_path, 
						 const unsigned int ie, const unsigned int il)const;
	Spectrum connect_paths(const Path<D> &eye_path, const Path<D> &light_path )const;
	void connect_paths(const Path<D> &eye_path, const Path<D> &light_path, 
						 std::list<RadianceSample> &samples, const Real delta_time, const unsigned int nb_time_samples)const;
public:
	BidirectionalPathTracing(World<D> *w, unsigned int incoming_samples = 1, FILE *_f_log = NULL, Film *_film = NULL)
		:Integrator<D>(w), m_incoming_samples(incoming_samples),
		m_inv_incoming_samples(1./static_cast<Real> (m_incoming_samples)), f_log(_f_log), film(_film),
		m_next_event_sampling_technique(NextEventSampling::MeanFreePath), m_angular_sampling_technique(AngularSampling::PhaseFunction),
		m_termination_sampling_technique(TerminationSampling::RussianRoulette), m_vertex_connection_technique(VertexConnection::ShadowConnection), 
		m_weighting_technique(WeightingTechnique::MIS_Power), m_beta_MIS(2.), m_avg_nb_vertices_path(4), m_max_path_size(30),m_path_trace_only(true),
		m_keep_track_sample_count(false)
	{
		set_mode(0);
	}
	~BidirectionalPathTracing()
	{}
	// ------------------------------------------------------------------------------------------
	void preprocess() 
	{
		if (f_log)
		{
			fprintf(f_log, "=============================================\n");
			fprintf(f_log, "Bidirectional Path Tracing\n");
			fprintf(f_log, "---------------------------------------------\n");
			fprintf(f_log, "# Samples: %d\n", m_incoming_samples);
			fprintf(f_log, "---------------------------------------------\n");
			fprintf(f_log, "-Next Event: ");
			switch(m_next_event_sampling_technique)
			{
			case NextEventSampling::Time:
				fprintf(f_log, "Time Sampling - #Vertices %d\n", m_avg_nb_vertices_path);
				break;
			case NextEventSampling::MeanFreePath:
			default:
				fprintf(f_log, "Mean Free Path Sampling\n");
			}
			fprintf(f_log, "-Angular Sampling: ");
			switch( m_angular_sampling_technique )
			{
			case AngularSampling::PhaseFunction:
				fprintf(f_log, "Phase Function\n");			
				break;
			case AngularSampling::ShortestTime:
				fprintf(f_log, "Shortest Time\n");
				break;
			}
			fprintf(f_log, "-Path Termination (Max Path Size %d): ", m_max_path_size );
			switch(m_termination_sampling_technique)
			{
			case TerminationSampling::TotalTime:
				fprintf(f_log, "Total Time\n");
				break;
			case TerminationSampling::RussianRoulette:
				fprintf(f_log, "Russian Roulette\n");
				break;
			}
			fprintf(f_log, "-Vertices Connection: ");
			switch( m_vertex_connection_technique )
			{
			case VertexConnection::TimeSampling:
				fprintf(f_log, "Point-to-Line\n");
				break;
			case VertexConnection::ShadowConnection:
				fprintf(f_log, "Shadow Connection\n");
				break;
			}
			fprintf(f_log, "=============================================\n");
		}
	}
	
	void set_mode(int mode)
	{
		switch( mode )
		{
		case 1:
			m_next_event_sampling_technique = NextEventSampling::Time;
			m_angular_sampling_technique = AngularSampling::ShortestTime;
			m_termination_sampling_technique = TerminationSampling::TotalTime;
			m_vertex_connection_technique = VertexConnection::TimeSampling;
			m_weighting_technique = WeightingTechnique::MIS_Balance;
			break;
		case 0:
		default:
			m_next_event_sampling_technique = NextEventSampling::MeanFreePath;
			m_angular_sampling_technique = AngularSampling::PhaseFunction;
			m_termination_sampling_technique = TerminationSampling::RussianRoulette;
			m_vertex_connection_technique = VertexConnection::ShadowConnection;
			m_weighting_technique = WeightingTechnique::MIS_Balance;
		};
	}
	void set_max_path_size( int max_path_size ){ m_max_path_size = max_path_size; }
	void set_path_tracing_only( bool path_trace_only) { m_path_trace_only = path_trace_only; }
	void keep_track_on_nb_samples( bool on = true ){ m_keep_track_sample_count = on; }
	// ------------------------------------------------------------------------------------------
	void operator()(const VectorN<D> &p, std::list<RadianceSample> &samples, const Real delta_time, const unsigned int nb_time_samples) const;

	void operator()(const Ray<D> &r, std::list<RadianceSample> &samples, const Real delta_time = 0., const unsigned int nb_time_samples = 1) const;
	Spectrum operator()(const Ray<D> &r) const;// {return Spectrum(0.);};
	
	void operator()(const Intersection<D> &it, std::list<RadianceSample> &samples, const Real delta_time = 0., const unsigned int nb_time_samples = 1) const;
	Spectrum operator()(const Intersection<D> &it) const;

}; //BidirectionalPathTracing

//=============================================================================
// Computes sampling distance. Note that here we assume that the ray has not been
// traced, and therefore the distance and pdf needs to be corrected aftewards
// (see function 'correct_next_event_distance').
template<int D>
Real BidirectionalPathTracing<D>::sample_next_event_distance(const Ray<D> &r, Real &p)const
{
	Real distance;
	Real lambda = r.get_medium()->get_extinction(r.get_origin()).avg();
	Real epsilon;

	Real total_time = film->get_time_resolution()*film->get_exposure_time();
	Real total_distance = total_time/m_world->time_of_flight(1)/m_world->get_ior();
	switch(m_next_event_sampling_technique)
	{
	case NextEventSampling::Time:
		lambda = m_avg_nb_vertices_path/total_distance;

	case NextEventSampling::MeanFreePath:
	default:
		{
			// Sample distance
			epsilon = RNG::StdRandom::get_real();
			distance = -logf(epsilon)/lambda;
			
			p = exp(-distance*lambda)*lambda;
		}
	}

	//printf("BDPT: Sampling distance: Lambda: %f, D: %f, P: %f\n", lambda, distance, p);

	return distance;
}

//=============================================================================
// NOTE: This function assumes that the ray 'r' has already
//		been traced, and that it is within participating media.
template<int D>
void BidirectionalPathTracing<D>::correct_next_event_distance( Ray<D> &r, Real &distance, 
																		   bool &in_media, Real &p )const
{
	if( !r.get_medium() )
	{
		distance = r.get_parameter();
		return;
	}

	Real lambda = r.get_medium()->get_extinction(r.get_origin()).avg();
	
	Real total_time = film->get_time_resolution()*film->get_exposure_time();
	Real total_distance = total_time/m_world->time_of_flight(1)/m_world->get_ior();
	switch(m_next_event_sampling_technique)
	{
	case NextEventSampling::Time:
		lambda = m_avg_nb_vertices_path/total_distance;
	
	case NextEventSampling::MeanFreePath:
	default:
		if( r.did_hit() && (r.get_parameter() < distance ) )
		{
			distance = r.get_parameter();
			in_media = false; 
			p = exp(-r.get_parameter()*lambda);

			//printf("BDPT: Correct Distance: Lambda: %f, D: %f, P: %f\n", lambda, distance, p);
		}
		else
		{
			r.set_parameter( distance );
			in_media = true;
		}
	}
}

//=============================================================================
// Sample next event's angular domain
template<int D>
void BidirectionalPathTracing<D>::sample_outgoing_ray( Ray<D> &ray, Intersection<D> &it, bool media_interaction, 
																   Real sampled_distance, Real current_delay, const VectorN<D> &target, 
																   Spectrum &f, Real &time, Real &p )const
{
	if( media_interaction )
	{
		// Random walk's next step
		VectorN<D> position = ray.get_position();
		VectorN<D> new_omega_i;
		Medium<D> *m = ray.get_medium();
		Spectrum u_t = m->get_extinction(position);
		Spectrum u_s = m->get_scattering(position);

		// Select sampling technique to use
		Real p_ts;
		switch( m_angular_sampling_technique )
		{
		case AngularSampling::PhaseFunction:
			p_ts = 0.;
			break;
		case AngularSampling::ShortestTime:
			p_ts = 1.;
			break;
		}
		
		// Create the angular time sampler
		// - Compute direction and distance to target
		VectorN<D> l = target-position;
		Real mod = l.length();
		// - Compute sampling boundaries
		Real ta = m_world->time_of_flight(mod)*m_world->get_ior() + 
			current_delay;
		Real tb = film->get_time_resolution()*film->get_exposure_time()+film->get_time_offset();

		// - And create the sampler
		Probability::TimeAngular<D> sa(ta, tb, sampled_distance, mod, 
									   m_world->get_ior(), 1./m_world->time_of_flight(1));

		// Check if between boundaries
		if( (tb - (ta+m_world->time_of_flight(sampled_distance)*m_world->get_ior())) < film->get_exposure_time() ) 
			p_ts = 0.;

		Real xi = RNG::StdRandom::get_real();
		// MIS between phase function sampling and angular time sampling
		if( xi< p_ts ) 
		{
			// Time sampling
			Real theta = sa.cdf_inv( RNG::StdRandom::get_real() );
			Real phi = 2*M_PI*RNG::StdRandom::get_real(); //Not sure about this
		
			VectorN<D> dir(cos(phi)*sin(theta), cos(theta), sin(phi)*sin(theta));
			new_omega_i = dir.transform_matrix_to(Vector3(0,1,0), l );


			// Compute f and p, with p = p/w using MIS
			f = u_s*m->f(position, ray.get_direction(), new_omega_i, time, p);

			p = m->p(position, ray.get_direction(), new_omega_i, time)*(1-p_ts)+
				sa.pdf(theta)/(2*M_PI)*p_ts;
			p /= p_ts;

			if( p < 1.e-5 ) p_ts = 0.;
		}

		if( xi >= p_ts)  // Phase Function Sampling
		{
			f = u_s*m->sample_direction(position, ray.get_direction(), new_omega_i, time, p);
			if( p_ts > 0 )
			{
				p = p*(1-p_ts)+sa.pdf(acos(dot(l,new_omega_i)))/(2*M_PI)*p_ts;
				p /= (1-p_ts);
			}
		}

		ray = Ray<D>(position, new_omega_i, ray.get_level()+1, ray.get_refraction_index(), ray.get_medium());
	}
	else //SURFACE HIT
	{
		//Random walk's next step			
		f = it.material()->sample_outgoing_ray(it, ray, time, p );
		f *= dot_abs(it.get_normal(), ray.get_direction()); 			
	}	
}


//=============================================================================
// Decide the termination of the random walk.
//	The function decides if the path finishes or not. The method to select the
//	termination is specified by 'm_termination_sampling_technique' from the fol-
//	lowing:
//		- TerminationSampling::TotalTime - Evaluates if the path total time is
//			longer than the total time of the film, and if it is, terminates the
//			random walk.
//		- WeightingTechnique::NbBounces - If the number of bounces of the path is
//			higher than the max the path is terminated.
//		- WeightingTechnique::RussianRoulette - Uses stochastic Russian roulette
//			to determine whether the path is terminated (absorbed) or not. This is
//			the only one unbiased.
//
template<int D>
bool BidirectionalPathTracing<D>::sample_termination( Real albedo, Real time, int nb_bounces, Real &p )const
{
	if( nb_bounces > m_max_path_size )
		return true;
	
	switch(m_termination_sampling_technique)
	{
	case TerminationSampling::NbBounces:
		if( nb_bounces > m_avg_nb_vertices_path )
			return true;
		
		p = 1.;
		break;

	case TerminationSampling::TotalTime:
		if( film )
		{
			if( time > ( film->get_time_length() + film->get_time_offset()) )
				return true;
			
			p = 1.;
			break;
		}
	case TerminationSampling::RussianRoulette:
	default:
		{
			Real epsilon = RNG::StdRandom::get_real();

			if (epsilon > albedo) 
				return true;

			p = albedo;
		}
	}

	return false;
}


//=============================================================================
// Compute path
template<int D>
void BidirectionalPathTracing<D>::generate_path(const Ray<D> &r, const Spectrum &f0, const Real p0, const VectorN<D> &target, Path<D> &path )const
{
	//Prepare variables for the loop
	Ray<D> curr_ray(r.get_origin(),r.get_direction(), false, r.get_level(), r.get_refraction_index(), r.get_medium()); 
	Intersection<D> curr_it;
	VectorN<D> position;
	
	Spectrum u_t, u_s, scat_albedo;
	Real avg_albedo;
	Real sampled_distance = std::numeric_limits<Real>::infinity(); //Sampled mean free path in free-space

	// Throughput and probability of the path
	Spectrum f = f0;
	Real p = 1., pi = p0;
	
	Real time = 0;

	// Variables used through the loop
	Spectrum fa(1.);
	Real pd(1.), pa(1.);
	Real ta = 0.; 


	bool media_interaction = false;
	bool first_iteration = true;

	// Iterate path
	while(1)
	{
		u_t = Spectrum(0.);
		u_s = Spectrum(0.);
		
		pd = 1.;
		// ------------------------------------------------------------------------------------------
		// First of all test if we are currently in a medium, and sample distance there...	
		if(curr_ray.get_medium())
		{	
			// If within media...
			u_t = curr_ray.get_medium()->get_extinction(curr_ray.get_origin());
			u_s = curr_ray.get_medium()->get_scattering(curr_ray.get_origin());
			
			// Sample Distance! 
			sampled_distance = sample_next_event_distance(curr_ray, pd);

			//media_interaction = true;
		} 

		// ------------------------------------------------------------------------------------------
		// With the sampled distance, sample the new ray
		fa = Spectrum(1.); pa=1.; ta=0.;
		if( first_iteration )
			first_iteration = false;
		else
			sample_outgoing_ray( curr_ray, curr_it, media_interaction, sampled_distance, time, target, fa, ta, pa );


		// ------------------------------------------------------------------------------------------
		// Trace next ray
		curr_it = Intersection<D>();
		this->m_world->first_intersection(curr_ray, curr_it);


		// Test if there's no scattering event in media or surfaces. If don't then the 
		// path is finished...
		if( !curr_ray.get_medium() && !curr_it.did_hit())
			return;

		// ------------------------------------------------------------------------------------------
		// And correct distances
		correct_next_event_distance( curr_ray, sampled_distance, media_interaction, pd );

		// ------------------------------------------------------------------------------------------
		// Update contribution, probability, accomulated probability, and time.
		f *= (fa*exp(-sampled_distance*u_t));
		pi *= (pd*pa);
		p *= pi; 
		time += ta + m_world->time_of_flight(sampled_distance)*curr_ray.get_refraction_index();;

		// ------------------------------------------------------------------------------------------
		// Store vertex
		if( media_interaction )
		{
			//return;
			position = curr_ray.get_origin() + curr_ray.get_direction()*sampled_distance;
			path.add_vertex(position, curr_ray.get_direction(), curr_ray.get_medium(), 
							f, p, pi, time); 
			
			scat_albedo = (u_s/u_t);
		}
		else
		{
			
			position = curr_it.get_position();
			if( !curr_it.material()->is_type(Reflectance::DELTA) )
				path.add_vertex(position, curr_ray.get_direction(), curr_it.get_normal(), curr_it.get_uv(), curr_it.material(), 
							f, p, pi, time); 
			
			scat_albedo = (1.-curr_it.material()->get_absorption(curr_it.get_uv()));
		}
		

		// ------------------------------------------------------------------------------------------
		// Finally, evaluate Sample Termination!
		pi = 1.f;
		avg_albedo = scat_albedo.avg();

		if( sample_termination( avg_albedo, time, curr_ray.get_level(), pi ))
			return;
	}
}

//=============================================================================
// Connect an eye vertex with a randomly sampled light
template<int D>
bool BidirectionalPathTracing<D>::connect_vertex_with_light( const typename Path<D>::Vertex *eye_vertex, 
																		 Spectrum &f, Real &p, Real *time )const
{
	LightSample<D> light_sample;
	Real pl1, pl2;

	if ( m_world->sample_light(pl1)->sample( eye_vertex->get_vertex_position(), light_sample, pl2 ))
	{	
		Spectrum fe;
		Real pe = 1., te = 0.;
		if( time )
		{
			eye_vertex->compute_scattering(-light_sample.dir, fe, te, pe);
			(*time)= m_world->time_of_flight(light_sample.dist)*m_world->get_ior() + te;
		}
		else
			eye_vertex->compute_scattering(-light_sample.dir, fe );

		// Compute attenuation if needed. Note that we are asuming that the 
		// medium is homogeneous, and the same over all the scene. This needs
		// to be fixed.
		Spectrum att(1.);
		if( m_world->get_medium()) 
			att = exp(-m_world->get_medium()->get_extinction(eye_vertex->get_vertex_position())*light_sample.dist);
		
		f = fe*light_sample.irradiance*att;

		p = pe*pl1*pl2;

		return true;
	}
	else
	{
		f = Spectrum();
		p = 1.;
		if( time )
			(*time) = m_world->time_of_flight(light_sample.dist)*m_world->get_ior();
		

		return false;
	}
}


//=============================================================================
// Connect two vertices
template<int D>
bool BidirectionalPathTracing<D>::connect_vertices( const typename Path<D>::Vertex *eye_vertex, const typename Path<D>::Vertex *light_vertex, 
																	Spectrum &f, Real *time, Real *pdf_time )const
{
	VectorN<D> l2e = eye_vertex->get_vertex_position()-light_vertex->get_vertex_position();
	Real d_l2e = l2e.length(); 
	l2e /= d_l2e;
	
	Ray<D> ray;

	if( time && (d_l2e+eye_vertex->get_subpath_delay()+light_vertex->get_subpath_delay()) > 
				 film->get_time_resolution()*film->get_exposure_time()+film->get_time_offset() )
		return false;


	switch( m_vertex_connection_technique )
	{
	//-----------------------------------------------------------------------------
	case VertexConnection::TimeSampling:
		if( m_world->get_medium() && time && film ) 
		{
			// Only makes sense if we are in a medium, and interested in time.
			// Otherwise, we use regular MFP sampling.
			Spectrum fo, ft, fv;
			Real po, pt, pv, psp, pd;
			Real to, tt, tv;

			const typename Path<D>::Vertex *o, *t, *v;

			// We first obtain the intermediate vertex
			// - Select if the origin is the light or the eye vertex.
			if( RNG::StdRandom::get_real() < .5 )
			{	o = eye_vertex; t = light_vertex; l2e = -l2e; }
			else
			{	o = light_vertex; t = eye_vertex;	}
			psp = .5;

			// - Sample a new direction
			VectorN<D> dir; 
			Real pdir;
			o->sample_direction(dir, fo, to, po);

			// - Sample distance using Time Sampling!
			Real dist=-1.;

			Real ta = m_world->time_of_flight(d_l2e)*m_world->get_ior() + 
				o->get_subpath_delay() + t->get_subpath_delay();

			Real tb = film->get_time_resolution()*film->get_exposure_time()+film->get_time_offset();
			if (ta >= tb ) return false;

			Probability::TimeLineToPoint<D> st(ta, tb, dir, l2e,  m_world->get_ior(), 1./m_world->time_of_flight(1));
			while(dist<0) dist = st.cdf_inv(RNG::StdRandom::get_real());
			pd = st.pdf(dist);

			// - Check if hitting a surface
			Intersection<D> it;
			ray = Ray<D>(o->get_vertex_position(), dir, true);
			ray.shift();

			bool media_interaction = true;
			m_world->first_intersection(ray, it);

			if( ray.did_hit() && (ray.get_parameter() < dist ) )
			{
				dist = ray.get_parameter();
				pd = st.cdf(dist);
				media_interaction = false;
			}

			// - Create the new vertex
			if( media_interaction)					
				v = new Path<D>::VolumeVertex(o->get_vertex_position() + dir*dist, 
							dir, m_world->get_medium(), Spectrum(1), 1.,1.,0.); 
			else
			{
				if( !it.material()->is_type(Reflectance::DELTA) )
					v = new Path<D>::SurfaceVertex(it.get_position(), dir, it.get_normal(), 
							it.get_uv(), it.material(), Spectrum(1), 1.,1.,0.);
				else
					return false;
			}		

			// Perform shadow connection
			VectorN<D> v2r = t->get_vertex_position() - v->get_vertex_position();
			Real d_v2r = v2r.length();
			v2r /= d_v2r;

			ray = Ray<D>(t->get_vertex_position(),v2r, false);
			ray.shift();
			if( m_world->intersects(ray,d_v2r) )
			{

				v->compute_scattering( v2r, fv, tv, pv);
				t->compute_scattering(-v2r, ft, tt, pt);
					
				(*time)= m_world->time_of_flight(dist)*m_world->get_ior()
						 + m_world->time_of_flight(d_v2r)*m_world->get_ior()
						 + to + tv + tt;
					
				(*pdf_time) = psp*po*pd*pv*pt;
				
				Spectrum att = exp(-m_world->get_medium()->get_extinction(light_vertex->get_vertex_position())*(d_v2r+dist));

				f = fo*fv*ft*att/(D==2?d_v2r:(d_v2r*d_v2r));

				delete v;
				return true;
			}

			// else...
			f = Spectrum();
			(*time) = m_world->time_of_flight(d_l2e)*m_world->get_ior();
			(*pdf_time) = 1.;
			delete v;
			
			return false;
		}
		
	//-----------------------------------------------------------------------------
	// Regular shadow connection
	case VertexConnection::ShadowConnection:
	default:
		{
			Spectrum fe, fl;
			Real pe, pl;
			Real te, tl;

			ray = Ray<D>(light_vertex->get_vertex_position(),l2e, false);
			ray.shift();
			if( m_world->intersects(ray,d_l2e) )
			{
				if( time )
				{
					eye_vertex->compute_scattering(-l2e, fe, te, pe);
					light_vertex->compute_scattering(l2e, fl, tl, pl);
					
					(*time)= m_world->time_of_flight(d_l2e)*m_world->get_ior() 
							 + te + tl;
					
					(*pdf_time) = pe*pl;
				}
				else
				{
					eye_vertex->compute_scattering( -l2e, fe );
					light_vertex->compute_scattering( l2e, fl );
				}

				// Compute attenuation if needed. Note that we are asuming that the 
				// medium is homogeneous, and the same over all the scene. This needs
				// to be fixed.
				Spectrum att(1.);
				if( m_world->get_medium()) 
					att = exp(-m_world->get_medium()->get_extinction(light_vertex->get_vertex_position())*d_l2e);

				f = fe*fl*att/(D==2?d_l2e:(d_l2e*d_l2e));
				return true;
			}
			else
			{
				f = Spectrum();
				if( time )
				{
					(*time) = m_world->time_of_flight(d_l2e)*m_world->get_ior();
					(*pdf_time) = 1.;
				}

				return false;
			}
		}
	}
}

//=============================================================================
// Compute the weight of the path.
//	The function weights the contribution of the path resulting on connecting
//	the subpaths eye_path[ie] and light_path[il], using the technique defined 
//	in 'm_weighting_technique', which can be:
//		- WeightingTechnique::Uniform - It applies a uniform weight to all
//			paths (w=1).
//		- WeightingTechnique::Balance - Computes the weight using MIS, using
//			the balance heuristic [1], as described in Veach's thesis [2].
//		- WeightingTechnique::Power - Computes the weight using MIS with the
//			the power heuristic [1,2], using the parameter 'm_beta_MIS'.
//
//   [1] E. Veach and L. Guibas. 1995. Optimally Combining Sampling Techniques 
//		for Monte Carlo Rendering. In SIGGRAPH '95.
//      http://www-graphics.stanford.edu/papers/combine/
//   [2] E. Veach. 1997. Robust Monte Carlo Methods for Light Transport
//      Simulation. PhD Thesis (Stanford University).
//      http://graphics.stanford.edu/papers/veach_thesis/
//
template<int D>
Real BidirectionalPathTracing<D>::compute_weights(const Path<D> &eye_path, const Path<D> &light_path, 
															   const unsigned int ie, const unsigned int il)const
{
	int beta = m_beta_MIS;

	switch(m_weighting_technique)
	{
	//------------------------------------------------------------------------
	// MIS
	case WeightingTechnique::MIS_Balance:
		beta = 1;
	case WeightingTechnique::MIS_Power:
		return PathMIS::compute_weight(eye_path, light_path, ie, il, beta);
			
		break;

	//------------------------------------------------------------------------
	// Uniform
	case WeightingTechnique::Uniform:
	default:		
		return 1.;
		break;
		
	}
}

// Connect paths in steady state
template<int D>
Spectrum BidirectionalPathTracing<D>::connect_paths(const Path<D> &eye_path, const Path<D> &light_path )const
{
	Spectrum radiance(0.);
	Spectrum f;
	Real p;
	Real w;

	int nb_eye_vertices = std::min(2,eye_path.size());
	// Iterate over all eye vertices
	//for( unsigned int ie=0; ie<nb_eye_vertices; ++ie )
	for( unsigned int ie=0; ie<eye_path.size(); ++ie )
	{
		// First vertex of light path is the light source.
		// Following [Veach 95] we avoid correlation by randomly sampling the direct light
		// for each vertex of the eye path.
		if( (m_components & Scattering::Level::SINGLE || ie>0) && connect_vertex_with_light( eye_path[ie], f, p ) )
		{

			f *= eye_path[ie]->get_vertex_value();
			p *= eye_path[ie]->get_subpath_pdf();
			
			radiance += f/p;
		}
		

		if( !(m_components & Scattering::Level::MULTIPLE) )
			break;

		// Iterate over all light vertices
		for( unsigned int il=0; il<light_path.size(); ++il )
		{
			f = Spectrum(0.);
			if( connect_vertices( eye_path[ie], light_path[il], f ) )
			{
				f *= (eye_path[ie]->get_vertex_value()) * (light_path[il]->get_vertex_value());
				p =  (eye_path[ie]->get_subpath_pdf()) * (light_path[il]->get_subpath_pdf());
				w = compute_weights(eye_path,light_path,ie,il);

				radiance += f*(w/p);
			}
		}
	}

	return radiance;
}

// Connect paths in transient state
template<int D>
void BidirectionalPathTracing<D>::connect_paths(const Path<D> &eye_path, const Path<D> &light_path, 
															 std::list<RadianceSample> &samples, const Real delta_time, const unsigned int nb_time_samples)const
{
	Spectrum f;
	Real p;
	Real w;
	Real t;

	Real total_time = film->get_time_resolution()*film->get_exposure_time()+film->get_time_offset();
	
	// Iterate over all eye vertices
	for( unsigned int ie=0; ie<eye_path.size(); ++ie )
	{
		// If the current vertex delay is longer than the film, all its events
		// and these in the subsequent vertices of the path will happen outside
		// the sensor window.
		if( eye_path[ie]->get_subpath_delay() > total_time )
			return;

		if( (ie+1) > m_max_path_size )
			return;

		// First vertex of light path is the light source.
		// Following [Veach 95] we avoid correlation by randomly sampling the direct light
		// for each vertex of the eye path.
		if( (m_components & Scattering::Level::SINGLE || ie>0) && connect_vertex_with_light( eye_path[ie], f, p, &t ) )
		{
			f *= eye_path[ie]->get_vertex_value();
			p *= eye_path[ie]->get_subpath_pdf();
			t += eye_path[ie]->get_subpath_delay();

			samples.push_back(RadianceSample(f / p*m_inv_incoming_samples, t, ie + 1));
		}
		else if( m_keep_track_sample_count )
			samples.push_back( RadianceSample(0., eye_path[ie]->get_subpath_delay()+t ));
		
		if( !(m_components & Scattering::Level::MULTIPLE) )
			break;

		// Iterate over all light vertices
		for( unsigned int il=0; il<light_path.size(); ++il )
		{
			if( (ie+il+2) > m_max_path_size )
				break;
			
			// Check if within the animation...	
			if( (eye_path[ie]->get_subpath_delay() + light_path[il]->get_subpath_delay()) > total_time )
				break;

			if( connect_vertices( eye_path[ie], light_path[il], f, &t, &p) )
			{
				f *= eye_path[ie]->get_vertex_value()*light_path[il]->get_vertex_value();
				p =  eye_path[ie]->get_subpath_pdf()*light_path[il]->get_subpath_pdf();
				t += eye_path[ie]->get_subpath_delay() + light_path[il]->get_subpath_delay();
				w = compute_weights(eye_path,light_path,ie,il);
				samples.push_back(RadianceSample(f / (p*w)*m_inv_incoming_samples, t, 2 + ie + il));
			}
			else if( m_keep_track_sample_count )
				samples.push_back( RadianceSample(0., 
					eye_path[ie]->get_subpath_delay() + light_path[il]->get_subpath_delay() + t ));
		}
	}
}

//=============================================================================
// Computes the light transport from an initial camera-sampled ray r.
template<int D>
Spectrum BidirectionalPathTracing<D>::operator()(const Ray<D> &r) const
{
	Spectrum reflected_radiance(0.);
	
	for(int sample = 0; sample < this->m_incoming_samples; ++sample )
	{
		Path<D> eye_path, light_path;

		// Sample light source. We do this first since it can be used
		// to guide the eye path when:
		// m_angular_sampling_technique = AngularSampling::ShortestTime.
		LightSample<D> light_sample;
		Real pl1, pl2;
		m_world->sample_light(pl1)->sample( light_sample, pl2 );
		
		// Generate eye path
		generate_path(r,Spectrum(1),1,light_sample.pos, eye_path);
		if( eye_path.size() == 0 )
			continue;

		// Generate light path
		if( !m_path_trace_only )
			generate_path(Ray<D>(light_sample.pos, light_sample.dir, false, 0, m_world->get_ior(), m_world->get_medium()),
						light_sample.irradiance*m_inv_incoming_samples, pl1*pl2, 
						r.get_origin(), light_path);
		
		// Connect vertices
		reflected_radiance += connect_paths(eye_path, light_path);		
	}
	//printf("Reflected radiance: [%f %f %f] vs %f\n", reflected_radiance[0],reflected_radiance[1],reflected_radiance[2],m_inv_incoming_samples);
	return reflected_radiance*m_inv_incoming_samples;
}


template<int D>
Spectrum BidirectionalPathTracing<D>::operator()(const Intersection<D> &it) const
{
	return (*this)(it.get_ray());
	
}

template<int D>
void BidirectionalPathTracing<D>::operator()(const Ray<D> &r, std::list<RadianceSample> &samples, const Real delta_time, const unsigned int nb_time_samples) const
{
	Spectrum reflected_radiance(0.);
	
	for(int sample = 0; sample < this->m_incoming_samples; ++sample )
	{
		Path<D> eye_path, light_path;

		// Sample light source. We do this first since it can be used
		// to guide the eye path when:
		// m_angular_sampling_technique = AngularSampling::ShortestTime.
		LightSample<D> light_sample;
		Real pl1, pl2;
		m_world->sample_light(pl1)->sample( light_sample, pl2 );
		
		// Generate eye path
		generate_path(r,Spectrum(1),1,light_sample.pos, eye_path);
		if( eye_path.size() == 0 )
			continue;

		// Generate light path
		if( !m_path_trace_only )
			generate_path(Ray<D>(light_sample.pos, light_sample.dir, false, 0, m_world->get_ior(), m_world->get_medium()),
						light_sample.irradiance, pl1*pl2, 
						r.get_origin(), light_path);
					
		// Connect vertices
		connect_paths(eye_path, light_path, samples, delta_time, nb_time_samples );	

		// Clear paths
		eye_path.clear();
		light_path.clear();
	}
}




template<int D>
void BidirectionalPathTracing<D>::operator()(const Intersection<D> &it, std::list<RadianceSample> &samples, const Real delta_time, const unsigned int nb_time_samples) const
{
	(*this)(it.get_ray(), samples, delta_time, nb_time_samples);
}

template<int D>
void BidirectionalPathTracing<D>::operator()(const VectorN<D> &p, std::list<RadianceSample> &samples, const Real delta_time, const unsigned int nb_time_samples) const
{
	for(int sample = 0; sample < this->m_incoming_samples; ++sample )
	{
		Path<D> eye_path, light_path;
		
		// Sample light source. We do this first since it can be used
		// to guide the eye path when:
		// m_angular_sampling_technique = AngularSampling::ShortestTime.
		LightSample<D> light_sample;
		Real pl1, pl2;
		m_world->sample_light(pl1)->sample( light_sample, pl2 );

		// Generate eye path
		// - sample eye ray
		VectorN<D> dir_ray;
		Real pdf_ray;
		Sampling::direction_uniformly(dir_ray,pdf_ray);
		// - ...and generate the path
		generate_path(Ray<D>(p, dir_ray, false, , 0, m_world->get_ior(), m_world->get_medium()), Spectrum(m_inv_incoming_samples),
						pdf_ray, light_sample.pos, eye_path);
		if( eye_path.size() == 0 )
			continue;

		// Generate light path
		generate_path(Ray<D>(light_sample.pos, light_sample.dir, false, 0, m_world->get_ior(), m_world->get_medium()),
						light_sample.irradiance*m_inv_incoming_samples, pl1*pl2, 
						p, light_path);
			
		
		// Connect vertices
		connect_paths(eye_path, light_path, samples, delta_time, nb_time_samples );		

		// Clear paths
		eye_path.clear();
		light_path.clear();
	}
}

#endif //_BIDIRECTIONAL_PATH_TRACING_INTEGRATOR_H_