/* 
* Copyright (C) 2015, Adrian Jarabo (http://giga.cps.unizar.es/~ajarabo/)
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the "Software"),
* to deal in the Software without restriction, including without limitation
* the rights to use, copy, modify, merge, publish, distribute, sublicense,
* and/or sell copies of the Software, and to permit persons to whom
* the Software is furnished to do so, subject to the following conditions:

* The above copyright notice and this permission notice shall be included
* in all copies or substantial portions of the Software.

* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
* DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
* OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef _PATH_H_
#define _PATH_H_

#include "Utils/globals.h"
#include "LinearAlgebra/VectorN.h"

using namespace std;

template<int D>
class Path
{
public:
	class Vertex
	{
	protected:
		VectorN<D> m_position;
		VectorN<D> m_direction;

		Spectrum m_f;
		Real m_p;
		Real m_pi;

		Real m_time, m_distance;

	public:
		Vertex(	const VectorN<D> &position, const VectorN<D> &direction, const Spectrum &f, Real p, Real pi)
			:m_position(position), m_direction(direction), m_f(f), m_p(p), m_pi(pi), m_time(0.), m_distance(0.){}

		Vertex(	const VectorN<D> &position, const VectorN<D> &direction, const Spectrum &f, 
				Real p, Real pi, Real time, Real distance)
			:m_position(position), m_direction(direction), m_f(f), m_p(p), m_pi(pi), m_time(time), m_distance(distance){}


		virtual void compute_scattering(const VectorN<D> &dir1, Spectrum &f1) const = 0;
		virtual void compute_scattering(const VectorN<D> &dir1, Spectrum &f1, Real &t1, Real &p1) const = 0;
		
		virtual void compute_probability(const VectorN<D> &dir1, Real &p1) const = 0;
		virtual void compute_probability(const VectorN<D> &dir1, Real t1, Real &p1) const = 0;
		

		virtual void sample_direction( VectorN<D> &dir1, Spectrum &f1, Real &p1 ) const = 0;
		virtual void sample_direction( VectorN<D> &dir1, Spectrum &f1, Real &t1, Real &p1 ) const = 0;

		inline const VectorN<D> &get_vertex_position()const{return m_position;}
		inline const Spectrum &get_vertex_value()const{return m_f;}
		inline Real get_vertex_pdf()const{return m_pi;}
		inline Real get_subpath_pdf()const{return m_p;}

		inline Real get_subpath_distance()const{return m_distance;}
		inline Real get_subpath_delay()const{return m_time;}

		inline Real operator-(const Vertex &v1){return (v1.m_position-this->m_position).length();}
	}; // Vertex

	class SurfaceVertex: public Vertex
	{
		VectorN<D> m_normal;
		VectorN<D-1> m_uv;
		const Material<D> *m_material;

	public:
		SurfaceVertex(	const VectorN<D> &position, const VectorN<D> &direction, const VectorN<D> &normal,
						const VectorN<D-1> &uv, const Material<D> *material,
						const Spectrum &f, Real p, Real pi, Real time = 0, Real distance=0)
			:Vertex(position, direction, f, p, pi, time, distance), m_normal(normal), m_uv(uv), m_material(material)
		{}

		void compute_scattering(const VectorN<D> &dir1, Spectrum &f1)const
		{
			f1 = m_material->f(m_direction, dir1, m_normal, m_uv)*dot_clamped(dir1,m_normal); //TODO: Check the sign of dir1!
		}
		
		void compute_scattering(const VectorN<D> &dir1, Spectrum &f1, Real &t1, Real &p1)const
		{
			f1 = m_material->f(Spectrum(1), m_direction, dir1, m_normal, m_uv, t1, p1)*dot_clamped(dir1,m_normal); //TODO: Check the sign of dir1!
		}

		void compute_probability(const VectorN<D> &dir1, Real &p1) const
		{
			p1 = m_material->p(m_direction, dir1, m_normal, m_uv);
		}

		void compute_probability(const VectorN<D> &dir1, Real t1, Real &p1)const
		{
			p1 = m_material->p(m_direction, dir1, m_normal, m_uv, t1);
		}
		
		void sample_direction( VectorN<D> &dir1, Spectrum &f1, Real &p1 ) const
		{
			f1 = m_material->sample_direction(m_direction, dir1, m_normal, m_uv, p1);
		}
		void sample_direction( VectorN<D> &dir1, Spectrum &f1, Real &t1, Real &p1 ) const
		{
			f1 = m_material->sample_direction(m_direction, dir1, m_normal, m_uv, t1, p1); //TODO: Check the sign of dir1!
		}

	}; // SurfaceVertex

	class VolumeVertex: public Vertex
	{
		Spectrum m_sigma_s;
		const Medium<D> *m_medium;
	public:
		VolumeVertex(	const VectorN<D> &position, const VectorN<D> &direction, const Medium<D> *medium,
						const Spectrum &f, Real p, Real pi, Real time = 0., Real distance = 0.)
			:Vertex(position, direction, f, p, pi, time, distance), m_medium(medium), 
			 m_sigma_s(medium->get_scattering(position))
		{}

		void compute_scattering(const VectorN<D> &dir1, Spectrum &f1)const
		{
			f1 = m_medium->f(m_position, m_direction, dir1)*m_sigma_s;//TODO: Check the sign of dir1!
		}
		
		void compute_scattering(const VectorN<D> &dir1, Spectrum &f1, Real &t1, Real &p1)const
		{
			f1 = m_medium->f(m_position, m_direction, dir1, t1, p1)*m_sigma_s; //TODO: Check the sign of dir1!
		}

		void compute_probability(const VectorN<D> &dir1, Real &p1) const
		{
			p1 = m_medium->p(m_position, m_direction, dir1);
		}

		void compute_probability(const VectorN<D> &dir1, Real t1, Real &p1)const
		{
			p1 = m_medium->p(m_position, m_direction, dir1, t1);
		}

		void sample_direction( VectorN<D> &dir1, Spectrum &f1, Real &p1 ) const
		{
			f1 = m_medium->sample_direction(m_position, m_direction, dir1, p1);
		}

		void sample_direction( VectorN<D> &dir1, Spectrum &f1, Real &t1, Real &p1 ) const
		{
			f1 = m_medium->sample_direction(m_position, m_direction, dir1, t1, p1); //TODO: Check the sign of dir1!
		}


	}; // VolumeVertex

private:
	vector<Vertex*> m_vertices;
	vector<SurfaceVertex> m_surface_vertices;
	vector<VolumeVertex> m_volume_vertices;

public:
	Path(){}
	~Path()
	{
		clear();
	}

	inline int size()const{return m_vertices.size();}
	
	inline const Vertex* operator[](unsigned int i)const
	{
		if( i>m_vertices.size() )
			return 0;

		return m_vertices[i];
	}


	inline void add_vertex(const VectorN<D> &position, const VectorN<D> &direction, 
					const VectorN<D> &normal, const VectorN<D-1> &uv, const Material<D> *material,
					const Spectrum &f, Real p, Real pi, Real time = 0, Real distance = 0)
	{
		m_vertices.push_back((Vertex*)(new SurfaceVertex(position, direction, normal, uv, material,
														 f, p, pi, time, distance)));
	}


	inline void add_vertex(const VectorN<D> &position, const VectorN<D> &direction, const Medium<D> *medium,
					const Spectrum &f, Real p, Real pi, Real time = 0, Real distance = 0)
	{
		m_vertices.push_back((Vertex*)(new VolumeVertex(position, direction, medium,
														 f, p, pi, time, distance)));
	}


	void clear()
	{
		vector<Vertex*>::iterator vx;
		for( vx=m_vertices.begin(); vx!=m_vertices.end(); (++vx))
			delete(*vx);

		m_vertices.clear();
	}
}; //Path

#endif //_PATH_H_