/* 
* Copyright (C) 2015, Adrian Jarabo (http://giga.cps.unizar.es/~ajarabo/)
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the "Software"),
* to deal in the Software without restriction, including without limitation
* the rights to use, copy, modify, merge, publish, distribute, sublicense,
* and/or sell copies of the Software, and to permit persons to whom
* the Software is furnished to do so, subject to the following conditions:

* The above copyright notice and this permission notice shall be included
* in all copies or substantial portions of the Software.

* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
* DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
* OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef _LAMBERTIAN_TEXTURED_H_
#define _LAMBERTIAN_TEXTURED_H_

#include "Material/Material.h"
#include "Material/Reflectance/Lambertian.h"
#include "image/image.h"

#include <vector>

using namespace std;

/**	Class defining the simplest emissing
	material. 
	TODO: Parametrize the light distribution,
		now it's only gaussian. */
template<int D>
class LambertianTextured: public Material<D>
{
	Spectrum m_kd;
	Imaging::Image *m_map;
public:
	LambertianTextured(): Material<D>(Reflectance::DIFFUSE),m_kd(1.),m_map(0){}
	LambertianTextured(const char* name):Material<D>(Reflectance::DIFFUSE),m_kd(1.),m_map(0)
	{
		m_map = new Imaging::Image(name);
	}
	LambertianTextured( Spectrum kd, const char* name ): Material<D>(Reflectance::DIFFUSE),m_kd(kd),m_map(0)
	{
		m_map = new Imaging::Image(name);
	}
	~LambertianTextured()
	{	if(m_map) delete m_map;	}

	
	Spectrum f(const VectorN<D> &omega_o, const VectorN<D> &omega_i, const VectorN<D> &normal, const Vector2 &uv) const
	{
		Spectrum kd = m_kd;

		if( m_map )
		{
			Imaging::RGBColor rgb = (*m_map)(m_map->width()*uv[0],m_map->height()*uv[1]);
			kd *= Spectrum(rgb.r(), rgb.g(), rgb.b());
		}
		
		Lambertian<D> reflectance(kd); 

		return reflectance.f(omega_o,omega_i,normal,uv);
	}
	Real p(const VectorN<D> &omega_o, const VectorN<D> &omega_i, const VectorN<D> &normal, const Vector2 &uv)const
	{
		Lambertian<D> reflectance(1.);

		return reflectance.p(omega_o,omega_i,normal,uv);
	}
	Spectrum sample_direction(const VectorN<D> &omega_o, VectorN<D> &omega_i, const VectorN<D> &normal, const Vector2 &uv, Real &pdf) const
	{
		Spectrum kd = m_kd;

		if( m_map )
		{
			Imaging::RGBColor rgb = (*m_map)(m_map->width()*uv[0],m_map->height()*uv[1]);
			kd *= Spectrum(rgb.r(), rgb.g(), rgb.b());
		}
		
		Lambertian<D> reflectance(kd); 

		return reflectance.sample_direction(omega_o,omega_i,normal,uv,pdf);
	}
	Spectrum sample_outgoing_ray(const Intersection<D> &it, Ray<D> &new_ray, Real &pdf) const
	{
		Spectrum kd = m_kd;

		if( m_map )
		{
			Imaging::RGBColor rgb = (*m_map)(m_map->width()*it.get_uv()[0],m_map->height()*it.get_uv()[1]);
			kd *= Spectrum(rgb.r(), rgb.g(), rgb.b());
		}
		
		Lambertian<D> reflectance(kd); 

		return reflectance.sample_outgoing_ray(it,new_ray,pdf);
	}

	// TIME RESOLVED MATERIAL FUNCTIONS
	// NOTE: to include re-radiation it's included as parameter the incoming radiance. This is pure crap, but without a reradiation matrix
	// as output (which I have no time now to do) is the only way to include re-radiation. This creates changes in "Path-tracing". But that's, as I said, wrong.
	// In fact, it'd need to adjust path-tracing to be purely recursive, since we cannot trace importance anymore.
	/** Computes the local direct reflected (or refracted) light. 
		omega_o is the view (incoming) direction, while omega_i is light (outgoing)*/
	virtual Spectrum f( const Spectrum &L_i, const VectorN<D> &omega_o, const VectorN<D> &omega_i, const VectorN<D> &normal, const Vector2 &uv, Real &delta_time, Real &pdf ) const 
	{	
		pdf = 1; delta_time = 0;
		return f(omega_o,omega_i,normal,uv)*L_i;
	}


	Spectrum get_absorption( const Vector2 &uv ) const
	{
		Spectrum kd = m_kd;

		if( m_map )
		{
			Imaging::RGBColor rgb = (*m_map)(m_map->width()*uv[0],m_map->height()*uv[1]);
			kd *= Spectrum(rgb.r(), rgb.g(), rgb.b());
		}
		
		return 1.-kd;
	}
}; //Emission

typedef LambertianTextured<3> LambertianTextured3D;
typedef LambertianTextured<2> LambertianTextured2D;

#endif //_EMISSION_H_