/* 
* Copyright (C) 2015, Adrian Jarabo (http://giga.cps.unizar.es/~ajarabo/)
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the "Software"),
* to deal in the Software without restriction, including without limitation
* the rights to use, copy, modify, merge, publish, distribute, sublicense,
* and/or sell copies of the Software, and to permit persons to whom
* the Software is furnished to do so, subject to the following conditions:

* The above copyright notice and this permission notice shall be included
* in all copies or substantial portions of the Software.

* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
* DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
* OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef _FRESNEL_DIELECTRIC_H_
#define _FRESNEL_DIELECTRIC_H_

#include "Material/Reflectance/BSDF.h"
#include "Material/Reflectance/Fresnel.h"
#include "RayTracing/Ray.h"
#include "RayTracing/Intersection.h"


template<int D>
class FresnelDielectric: public BSDF<D>
{
	//Absorption coefficient
	Spectrum m_absorption;

	//Index of refraction
	Real m_n, m_k;
	Real m_n2, m_k2;

public:
	FresnelDielectric():m_absorption(0.f), m_n(1.f), m_k(0.), m_n(1.f), m_k(0.), BSDF<D>( Reflectance::DELTA_TRANSMISSION ){}
	FresnelDielectric(const Spectrum &absorption, const Real ior, const Real ior_complex = 0.f)
		:m_absorption(absorption), m_n(ior), m_k(ior_complex), m_n2(ior*ior), m_k2(ior_complex*ior_complex), BSDF<D>(Reflectance::DELTA_TRANSMISSION){}
	FresnelDielectric(const Real ior, const Real ior_complex = 0.f)
		:m_absorption(0.f), m_n(ior), m_k(ior_complex), m_n2(ior*ior), m_k2(ior_complex*ior_complex), BSDF<D>(Reflectance::DELTA_TRANSMISSION){}
	
	Spectrum f(const VectorN<D> &omega_o, const VectorN<D> &omega_i, const VectorN<D> &normal, const Vector2 &uv) const
	{
		if( dot(omega_o.refract(normal, m_n), omega_i) > 0.99 )
			return (1.-m_absorption)/dot_abs(normal, omega_i);

		return Spectrum(0.);
	}
	Real p(const VectorN<D> &omega_o, const VectorN<D> &omega_i, const VectorN<D> &normal, const Vector2 &uv) const
	{
		return 0.;
	}
	Spectrum sample_direction(const VectorN<D> &omega_o, VectorN<D> &omega_i, const VectorN<D> &normal, const Vector2 &uv, Real &pdf) const
	{
		Real R = 0.;//Fresnel::fresnel_reflectance( omega_o, normal, m_n, m_k );

		//printf("R: %f\n", R);
		// Russian roulette
		Real epsilon1 = static_cast<Real>(rand())/static_cast<Real>(RAND_MAX);
		while( epsilon1 <= 0. )
			epsilon1 = static_cast<Real>(rand())/static_cast<Real>(RAND_MAX);
		
		//epsilon1 = 0.;		

		if (epsilon1 < R ) // Fresnel reflection
		{
			omega_i = omega_o.reflect(normal);
			pdf = R;
		}
		else // Refraction
		{
			pdf = 1-R; 
			omega_i = omega_o.refract(normal, m_n); omega_i.normalize();
		}

		Real aux_geom = dot_abs(normal, omega_i); aux_geom = (aux_geom<1.e-4)?1.:aux_geom;
		return (1.-m_absorption);///dot_abs(normal, omega_i);
//		Spectrum color = (R>.5)?Spectrum((R-.5)*2,1.-(R-.5)*2,0.):Spectrum(0.,R*2,1.-R*2);
//		return (1.-m_absorption)*color/dot_abs(normal, omega_i);

	}
	Spectrum sample_outgoing_ray(const Intersection<D> &it, Ray<D> &new_ray, Real &pdf) const
	{
		VectorN<D> omega;
		Real R = Fresnel::fresnel_reflectance( it.get_ray().get_direction(), it.get_normal(), m_n, m_k );
		Real ior;
		bool reflected = 1;

		// Russian roulette
		Real epsilon1 = static_cast<Real>(rand())/static_cast<Real>(RAND_MAX);
		while( epsilon1 <= 0. )
			epsilon1 = static_cast<Real>(rand())/static_cast<Real>(RAND_MAX);
		
		//epsilon1 = 0.;

		if (epsilon1 < R ) // Fresnel reflection
		{
			omega = it.get_ray().get_direction().reflect(it.get_normal());
			pdf = 1.;
			ior = it.get_ray().get_refraction_index();
		}
		else // Refraction
		{
			omega = it.get_ray().get_direction().refract(it.get_normal(), m_n);
			omega.normalize();
			pdf = 1.; 
			ior = (dot(omega, it.get_normal()) < 0.)?DEFAULT_REFRACTION_INDEX:m_n;

		}

		
		//Test if the ray enters the media or if it is leaving
		new_ray = Ray<D>(it.get_position(), omega, true,
						 it.get_ray().get_level()+1, ior, 
						 NULL);
		
		new_ray.shift();

		//Spectrum color = pdf>.5?Spectrum((pdf-.5)*2,1.-(pdf-.5)*2,0.):Spectrum(0.,pdf*2,1.-pdf*2);
		//return (1.-m_absorption)*color/dot_abs(it.get_normal(), new_ray.get_direction());

		return (1.-m_absorption);///dot_abs(it.get_normal(), new_ray.get_direction());
	}

	Spectrum get_absorption( const Vector2 &uv ) const{ return m_absorption; }
}; //FresnelDielectric

typedef FresnelDielectric<3> FresnelDielectric3D;
typedef FresnelDielectric<2> FresnelDielectric2D;

#endif //_TRANSMISSIVE_H_