#ifndef _WARD_H_
#define _WARD_H_

#include "Material/Reflectance/BSDF.h"
#include "RayTracing/Ray.h"
#include "RayTracing/Intersection.h"
#include <math.h>


/** Ward glossy BRDF, based on Bruce Walter's technical report:
	Walter, B. 2005. "Notes on the Ward BRDF"
	http://www.graphics.cornell.edu/~bjw/wardnotes.pdf
	[TODO] Develop 2D version of the BRDF, right now only 3D. */
class Ward: public BSDF<3>
{
	//Specular coefficients
	Spectrum Ps;
	//Anisotropic Gloss Lobe
	Real Ax, Ay;
public:
	Ward(const Spectrum &_Ps = Spectrum(.67)):Ps(_Ps),Ax(0.), Ay(0.),BSDF<3>(Reflectance::GLOSSY){}
	Ward(const Spectrum &_Ps, const Real _A ): Ps(_Ps),Ax(_A),Ay(_A),BSDF<3>(Reflectance::GLOSSY){}
	Ward(const Spectrum &_Ps, const Real _Ax, const Real _Ay): Ps(_Ps),Ax(_Ax),Ay(_Ay),BSDF<3>(Reflectance::GLOSSY){}
	
	Spectrum f(const Vector3 &omega_o, const Vector3 &omega_i, const Vector3 &normal, const Vector2 &uv)const
	{
		Real dot_o = dot(normal,omega_o), dot_i = dot(normal,-omega_i);
		if( dot_o < 0. || dot_i < 0. ) return Spectrum(0.);

		Vector3 tanx = cross(normal,Vector3(0.,0.,1.));
		Vector3 tany = cross(normal,tanx);
		Vector3 half_vector = -omega_i+omega_o; half_vector.normalize();
		
		Real hn = dot(half_vector,normal);
		Real hx = dot(half_vector,tanx)/Ax;
		Real hy = dot(half_vector,tany)/Ay;

		Spectrum left = Ps/(4*M_PI*Ax*Ay*sqrtf(dot_o*dot_i));
		Real right = exp(-(hx*hx + hy*hy)/(hn*hn));
		
		return left*right;
	}
	Spectrum sample_direction(const Vector3 &omega_o, Vector3 &omega_i, const Vector3 &normal, const Vector2 &uv, Real &pdf) const
	{
		//Sample hemisphere cosine weighted-> Need to be redone...
		Real epsilon1 = static_cast<Real>(rand())/static_cast<Real>(RAND_MAX);
		Real epsilon2 = static_cast<Real>(rand())/static_cast<Real>(RAND_MAX);
		
		while( epsilon2 == 0. )
			epsilon2 = static_cast<Real>(rand())/static_cast<Real>(RAND_MAX);

		
		Real phi = atan(Ay/Ax*atanf(2*M_PI*epsilon1));
		Real cos_phi = cos(phi)/Ax, sin_phi = sin(phi)/Ay;

		Real tan_theta = sqrtf(-logf(epsilon2)/(cos_phi*cos_phi+sin_phi*sin_phi));
		Real theta = tanf(tan_theta);
		Real sin_theta = sinf(theta), cos_theta = cosf(theta);
		Vector3 half_vector(cos_phi*sin_theta, cosf(theta), sin_phi*sin_theta);

		Real left = 1./(4*M_PI*Ax*Ay*dot(omega_o,half_vector)*cos_theta*cos_theta*cos_theta);
		Real right = exp(-tan_theta*tan_theta*(cos_phi*cos_phi+sin_phi*sin_phi));
		pdf = left*right;
		
		Vector3 dir = 2*dot(omega_o,half_vector)*half_vector - omega_o;

		omega_i = dir.transform_matrix_to(Vector3(0,1,0), normal);

		return f(-omega_o, -omega_i, normal, uv);
	}
	Spectrum sample_outgoing_ray(const Intersection3D &it, Ray3D &new_ray, Real &pdf) const
	{
		Vector3 new_omega_i;
		Spectrum refl = sample_direction(it.get_ray().get_direction(), new_omega_i, it.get_normal(),it.get_uv(), pdf);

		new_ray = Ray3D(it.get_position(), new_omega_i, true, it.get_ray().get_level()+1, 
			it.get_ray().get_refraction_index(), it.get_ray().get_medium());
		
		new_ray.shift();

		return refl;
	}

	Spectrum get_absorption( const Vector2 &uv ) const
	{	return Ps;	}
}; //Ward

#endif //_WARD_H_