/* 
* Copyright (C) 2015, Adrian Jarabo (http://giga.cps.unizar.es/~ajarabo/)
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the "Software"),
* to deal in the Software without restriction, including without limitation
* the rights to use, copy, modify, merge, publish, distribute, sublicense,
* and/or sell copies of the Software, and to permit persons to whom
* the Software is furnished to do so, subject to the following conditions:

* The above copyright notice and this permission notice shall be included
* in all copies or substantial portions of the Software.

* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
* DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
* OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef _UBER_MATERIAL_H_
#define _UBER_MATERIAL_H_

#include "Material/Material.h"
#include "RayTracing/Ray.h"
#include "RayTracing/Intersection.h"
#include "Material/Reflectance/BSDF.h"

#include <vector>

template<int D> class BSDF;

using namespace std;

template<int D>
class UberMaterial: public Material<D>
{
	std::vector<BSDF<D>*> bsdfs;
	std::vector<Real> weights;
	Real total_weight, inv_total_weight;
public:
	UberMaterial():total_weight(0.),inv_total_weight(0.){}
	void add_bsdf(BSDF<D>* bsdf, const Real weight = 1.f)
	{
		bsdfs.push_back(bsdf);
		weights.push_back(weight);
		total_weight += weight;
		inv_total_weight = 1./total_weight;

		m_type = static_cast<Reflectance::Type>(m_type | bsdf->get_type());
	}
	Spectrum f(const VectorN<D> &omega_o, const VectorN<D> &omega_i, const VectorN<D> &normal, const Vector2 &uv) const
	{
		Spectrum result;
		for( int i=0; i<bsdfs.size(); ++i)
			result += bsdfs[i]->f(omega_o, omega_i, normal, uv)*weights[i];

		return result * inv_total_weight;
	}
	Real p(const VectorN<D> &omega_o, const VectorN<D> &omega_i, const VectorN<D> &normal, const Vector2 &uv)const
	{
		Real pdf(0.);
		for( int i=0; i<bsdfs.size(); ++i)
			pdf += bsdfs[i]->p(omega_o, omega_i, normal, uv)*weights[i];

		return pdf * inv_total_weight;
	}
	
	Spectrum sample_direction(const VectorN<D> &omega_o, VectorN<D> &omega_i, const VectorN<D> &normal, const Vector2 &uv, Real &pdf) const
	{
		// Need for importance sampling...
		int idx = static_cast<Real>(rand())/static_cast<Real>(RAND_MAX+1)*(bsdfs.size());
		
		Spectrum F = bsdfs[idx]->sample_direction(omega_o, omega_i, normal, uv, pdf );
		pdf *= total_weight/weights[idx];
		
		return F;
	}
	Spectrum sample_outgoing_ray(const Intersection<D> &it, Ray<D> &new_ray, Real &pdf) const
	{
		new_ray = Ray<D>(this->m_default_n, this->m_default_medium);
		// Need for importance sampling...
		int idx = static_cast<Real>(rand())/static_cast<Real>(RAND_MAX+1)*bsdfs.size();
		
		Spectrum F = bsdfs[idx]->sample_outgoing_ray(it, new_ray, pdf );
		pdf *= total_weight/weights[idx];

		return F;
	}

	Spectrum get_absorption( const Vector2 &uv ) const
	{
		Spectrum result;
		for( int i=0; i<bsdfs.size(); ++i)
			result += bsdfs[i]->get_absorption(uv)*weights[i];

		return result * inv_total_weight;
	}
}; //UberMaterial

typedef UberMaterial<3> UberMaterial3D;
typedef UberMaterial<2> UberMaterial2D;

#endif //_UBER_MATERIAL_H_