/* 
* Copyright (C) 2015, Adrian Jarabo (http://giga.cps.unizar.es/~ajarabo/)
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the "Software"),
* to deal in the Software without restriction, including without limitation
* the rights to use, copy, modify, merge, publish, distribute, sublicense,
* and/or sell copies of the Software, and to permit persons to whom
* the Software is furnished to do so, subject to the following conditions:

* The above copyright notice and this permission notice shall be included
* in all copies or substantial portions of the Software.

* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
* DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
* OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef _HENYEY_GREENSTEIN_PHASE_FUNCTION_H_
#define _HENYEY_GREENSTEIN_PHASE_FUNCTION_H_

#include "Media/PhaseFunction/PhaseFunction.h"

namespace PhaseFunction
{
	//-------------------------------------------------------------------------------------------------
	// Henyey-Greenstein phase function. 
    template<int D>
	class HenyeyGreenstein: public PhaseFunction<D>
	{
		Real g, inv_2g;
	public:
		HenyeyGreenstein(const Real _g, const Real mean_time): 
	        PhaseFunction<D>(mean_time), g(_g), inv_2g(0.5/g) 
        {}
		//HenyeyGreenstein(const PhaseFunction<D> &pf): g(0)
		//{ /*TODO: Integrate phase function numerically on the sphere to obtain g*/}
				
		Spectrum sample_direction(const VectorN<D> &wi, VectorN<D> &wo, Real &pdf) const;
		Spectrum operator()(const VectorN<D>& wi, const VectorN<D>& wo) const;
		Real p(const VectorN<D>& wi, const VectorN<D> &wo) const;


		//TIME RESOLVED PHASE FUNCTIONS
        Spectrum sample_direction(const VectorN<D>& wi, VectorN<D> &wo, 
                                          Real &delta_time, Real &pdf) const
		{        
			return PhaseFunction<D>::sample_direction(wi,wo,delta_time,pdf);
		}

        virtual Spectrum operator()(const VectorN<D>& wi, const VectorN<D>& wo, 
                                    Real &delta_time, Real &pdf) const
        {
			return PhaseFunction<D>::operator()(wi,wo,delta_time,pdf);
		}

		virtual Real p(const VectorN<D>& wi, const VectorN<D> &wo, const Real delta_time)const
		{	
			return PhaseFunction<D>::p(wi,wo,delta_time);
		}
	}; // class HenyeyGreenstein


	template<>
	Spectrum HenyeyGreenstein<2>::sample_direction(const VectorN<2> &wi, VectorN<2> &wo, Real &pdf) const
	{
		Real epsilon1 = RNG::StdRandom::get_real();
		Real cos_theta = cos(std::min(2.*atan((1.-g)*tan(M_PI_2*(1.-2.*epsilon1))/(1.+g)), 1.));
		Real sin_theta = (rand() %2 == 1?1:-1)*sqrtf(std::max(0., 1.-cos_theta*cos_theta));
		
		Real HG = (1.-g*g)/(2.*M_PI*(1.+g*g-2.*g*cos_theta));
		pdf = HG;
		
		Vector2 dir(sin_theta, cos_theta);
		
		wo = dir.transform_matrix_to(Vector2(0,1), wi);
		
		return Spectrum(HG);
	}

	template<>
	Spectrum HenyeyGreenstein<3>::sample_direction(const Vector3& wi, Vector3 &wo, Real &pdf) const
	{
		Real epsilon1 = RNG::StdRandom::get_real();
		Real epsilon2 = RNG::StdRandom::get_real();
		
		Real phi = 2.*M_PI*epsilon2;
		
		Real aux = (1.-g*g)/(1-g+2.*g*epsilon1);
		
		Real cos_theta = std::min(inv_2g*(1.+g*g-aux*aux), 1.);
		Real sin_theta = sqrtf(std::max(0., 1.-cos_theta*cos_theta));
		
		Real HG = (1.-g*g)/(4.*M_PI*pow(1.+g*g-2.*g*cos_theta,1.5));
		pdf = HG;
		
		Vector3 dir(cos(phi)*sin_theta, cos_theta, sin(phi)*sin_theta);
		
		wo = dir.transform_matrix_to(Vector3(0,1,0), wi);
		
		return Spectrum(HG);
	}

	template<>
	Spectrum HenyeyGreenstein<2>::operator()(const Vector2& wi, const Vector2& wo) const
	{
		Real cos_theta = wi.dot(wo);
		return Spectrum((1.-g*g)/(2.*M_PI*(1.+g*g-2.*g*cos_theta)));
	}

	template<>
	Spectrum HenyeyGreenstein<3>::operator()(const Vector3& wi, const Vector3& wo) const
	{
		Real cos_theta = wi.dot(wo);
		return Spectrum((1.-g*g)/(4.*M_PI*pow(1.+g*g-2.*g*cos_theta,1.5)));
	}
    
	template<>
	Real HenyeyGreenstein<2>::p(const VectorN<2>& wi, const VectorN<2> &wo) const
	{
		Real cos_theta = dot(wi,wo);
		return (1.-g*g)/(2.*M_PI*(1.+g*g-2.*g*cos_theta));
	}

	template<>
	Real HenyeyGreenstein<3>::p(const VectorN<3>& wi, const VectorN<3> &wo) const
	{
		Real cos_theta = dot(wi,wo);
		return (1.-g*g)/(4.*M_PI*pow(1.+g*g-2.*g*cos_theta,1.5));
	}

}; //namespace PhaseFunction


#endif //_HENYEY_GREENSTEIN_PHASE_FUNCTION_H_