/* 
* Copyright (C) 2015, Adrian Jarabo (http://giga.cps.unizar.es/~ajarabo/)
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the "Software"),
* to deal in the Software without restriction, including without limitation
* the rights to use, copy, modify, merge, publish, distribute, sublicense,
* and/or sell copies of the Software, and to permit persons to whom
* the Software is furnished to do so, subject to the following conditions:

* The above copyright notice and this permission notice shall be included
* in all copies or substantial portions of the Software.

* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
* DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
* OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef _WORLD_H
#define _WORLD_H

#include <list>
#include <vector>
#include <assert.h>
#include "Utils/globals.h"
#include "Color/Spectrum.h"
#include "LightSource/LightSource.h"
#include "Material/Material.h"
#include "Integrator/Integrator.h"
#include "Geometry/ObjectPointer.h"
#include "Geometry/Aggregate/Aggregate.h"

template<int D> class Medium;

template<int D>
Medium<D> *Material<D>::m_default_medium=0;

template<int D>
Real Material<D>::m_default_n=1.;

/** This class contains all the information about the scene, including
	the geometry, materials, light sources, and background.		*/
template<int D>
class World
{
	/** Geometry in the World */
	Aggregate<ObjectPointer<D>,D> geometry;
	Real m_scene_radius;

	/** Materials in the World */

	/** Lights in the World */
	std::vector<LightSource<D>*> light_source_list;

	/** Background colour used to shade rays that miss all objects.
		It should probably something more programable... Maybe tem-
		platize it to support environment maps or things like that. */
	Spectrum background;

	/** Default world's medium and index of refraction 
		i.e. the medium and speed of light which is not defined by 
		any volume. */
	Medium<D> *medium;
	Real ior;
	//LambertianCamera* lcam;
public:
	World(): background(1.), medium(0), ior(1) 
	{	Material<D>::set_default_medium(0); Material<D>::set_default_refraction_index(1);	}
	
	//-----------------------------------------------------------
	// Set & get background color.
	void set_background(const Spectrum &bkg){ background = bkg;}
	Spectrum get_background( const VectorN<D> &omega ) const {return background;}

	//-----------------------------------------------------------
	// Light sources' functions
	inline void add_light(LightSource<D>* ls) 
	{
		light_source_list.push_back(ls);
	}
	const LightSource<D> *light(const int idx)const
	{
		return (light_source_list[idx]);
	}
	const LightSource<D> *sample_light(const Intersection<D> &it, Real &pdf)const
	{
		//Might be improved...
		return sample_light(pdf);
	}
	const LightSource<D> *sample_light(Real &pdf) const
	{
		//Might be improved...

		Real idx_f = static_cast<Real>(rand())/static_cast<Real>(RAND_MAX)*light_source_list.size();
		while(idx_f < 0. || idx_f >= 1.){
			idx_f = static_cast<Real>(rand())/static_cast<Real>(RAND_MAX)*light_source_list.size();
		}
		int idx = (int)floor(idx_f);
		pdf = 1./static_cast<Real>(light_source_list.size());

		return (light_source_list[idx]);
	}
	inline int nb_lights()const
	{
		return light_source_list.size();
	}
	

	//-----------------------------------------------------------
	// Geometry
	inline void add_object(const ObjectPointer<D> &o)
	{	geometry.add_primitive(o); }
	inline void add_objects(std::vector<ObjectPointer<D> > &os)
	{	geometry.add_primitives(os);	}

	inline void freeze()
	{	
		geometry.freeze(); 
		m_scene_radius = geometry.get_bounding_box().get_max_dimension()/2;
	}
	
	
	//-----------------------------------------------------------
	// Cameras
	/*inline void set_lambertian_camera (const LambertianCamera* _lcam)
	{	lcam = _lcam; }
	
	inline const LambertianCamera *get_lambertian_camera ()
	{	return lcam; }*/

	//-----------------------------------------------------------
	// World's Medium
	inline void set_medium( Medium<D> *m )
	{	medium = m;	Material<D>::set_default_medium(m); }
	inline void set_ior( Real n )
	{	ior = n; Material<D>::set_default_refraction_index(n);	}
	inline Medium<D>* get_medium() const
	{	return medium; }

	// Return the object that first intersects `ray'
	inline bool first_intersection(Ray<D>& ray, Intersection<D> &it, const Real &max_length = std::numeric_limits<Real>::infinity()) const
	{	return geometry.intersect(ray,it,max_length);	}

	inline bool intersects(Ray<D> &ray, const Real &max_length = std::numeric_limits<Real>::infinity(), const Real epsilon = _SIGMA_VISIBILITY_) const
	{	
		return geometry.intersect(ray, max_length - epsilon*m_scene_radius);;
	}
	
	inline bool is_visible(const VectorN<D> &v1, const VectorN<D> &v2)const
	{	
		return ~geometry.intersect(Ray<D>(v1,v2-v1,true),(v2-v1).length()); 
	}

	// Return Light travelling through the ray.
	// NOTE: Legacy code; this functions are deprecated.
	RadianceSample Li(Ray<D>& ray, Integrator<D> *integrator) const
	{
		// Define the default medium parameters in the ray.
		ray.set_medium(this->get_medium());
		ray.set_refraction_index(this->get_ior());

		
		// And intersect
		Intersection<D> it;
		first_intersection(ray, it);
		Spectrum radiance = (*integrator)(ray);

		
		return RadianceSample(radiance,0,0,it.get_ray().get_parameter());
	}
	void Li(Ray<D>& ray, Integrator<D> *integrator, std::list<RadianceSample> &samples, Real delta_time = 0., int nb_time_samples = 0.) const
	{
		// Define the default medium parameters in the ray.
		ray.set_medium(this->get_medium());
		ray.set_refraction_index(this->get_ior());

		(*integrator)(ray, samples, delta_time, nb_time_samples);			
	}
	
	void Li(VectorN<D>& p, Integrator<D> *integrator, std::list<RadianceSample> &samples, Real delta_time = 0., int nb_time_samples = 0.) const
	{
		//Computes radiance at a single point in the scene
		(*integrator)(p, samples, delta_time);
	}

	// Returns time of flight in pico seconds. Of course, this should be changed and improved
	Real time_of_flight(Real distance)const
	{	return distance / 0.000299792458f; }

	// IOR 4 VRL
	inline Real get_ior() const
	{
		return ior;
	}
}; //World

typedef World<3> World3D;
typedef World<2> World2D;

#endif //_WORLD_H




