/* 
* Copyright (C) 2015, Adrian Jarabo (http://giga.cps.unizar.es/~ajarabo/)
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the "Software"),
* to deal in the Software without restriction, including without limitation
* the rights to use, copy, modify, merge, publish, distribute, sublicense,
* and/or sell copies of the Software, and to permit persons to whom
* the Software is furnished to do so, subject to the following conditions:

* The above copyright notice and this permission notice shall be included
* in all copies or substantial portions of the Software.

* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
* DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
* OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef _SAMPLING_H_
#define _SAMPLING_H_

#include "Utils/globals.h"
#include "Utils/RandomNumbers.h"
#include <math.h>


template<class RNG>
class TSampling
{
public:
	static void jittered(Real &x, Real &pdf, Real min = 0., Real max = 1.)
	{
		x = RNG::get_real(); 
		x *= (max-min); x+= min;
		pdf = 1./(max-min);
	}
	static void jittered(Real &x, Real &y, Real &pdf, Real min = 0., Real max = 1.)
	{
		Real pdfx;
		jittered(x, pdfx, min, max);
		jittered(y, pdf, min, max);	
		pdf *= pdfx;
	}
	static void jittered(Real &x, Real &y, Real &z, Real &pdf, Real min = 0., Real max = 1.)
	{
		Real pdfx, pdfy;
		jittered(x, pdfx, min, max);
		jittered(y, pdfy, min, max);	
		jittered(z, pdf, min, max);
		pdf *= pdfx*pdfy;
	}
	
	static void circular(Vector2 &v, Real &pdf)
	{
		Real epsilon1 = RNG::get_real();
		Real phi = 2*M_PI*epsilon1;

		pdf = (.5*M_1_PI);
		
		v[0] = cos(phi); 
		v[1] = sin(phi);
	}
	static void spherical(Vector3 &v, Real &pdf)
	{	
		Real epsilon1 = RNG::get_real();
		Real epsilon2 = RNG::get_real();

		Real phi = 2*M_PI*epsilon1;
		Real cos_theta = 1-2*epsilon2;
		Real sin_theta = sqrt(1.-cos_theta*cos_theta);

		pdf = (.25*M_1_PI);
		
		v[0] = cos(phi)*sin_theta; 
		v[1] = epsilon2; 
		v[2] = sin(phi)*sin_theta; 
	}

	static inline void direction_uniformly(Vector2 &v, Real &pdf)
	{	circular(v, pdf);	}
	static inline void direction_uniformly(Vector3 &v, Real &pdf)
	{	spherical(v, pdf);	}
	
	static void hemispherical(Vector3 &v, Real &pdf)
	{
		Real epsilon1 = RNG::get_real();
		Real epsilon2 = RNG::get_real();

		Real phi = 2*M_PI*epsilon1;
		Real sin_theta = sqrt(1.-epsilon2*epsilon2);

		pdf = (.5*M_1_PI);
		
		v = Vector3(cos(phi)*sin_theta, fabs(epsilon2), sin(phi)*sin_theta);
	}

	static void cosine_weighted(Vector2 &v, Real &pdf)
	{
		Real epsilon = RNG::get_real();
		while( epsilon == 0. )
			epsilon = RNG::get_real();

		Real cos_theta = sqrt(epsilon);
		Real sin_theta = sqrt(1.-cos_theta*cos_theta);

		v = Vector2(sin_theta, cos_theta);

		pdf = (cos_theta*M_1_PI);
	}

	static void cosine_weighted(Vector3 &v, Real &pdf)
	{
		Real epsilon1 = RNG::get_real();
		Real epsilon2 = RNG::get_real();
		while( epsilon2 == 0. )
			epsilon2 = RNG::get_real();

		Real phi = 2*M_PI*epsilon1;
		Real cos_theta = sqrt(epsilon2);
		Real sin_theta = sqrt(1.-cos_theta*cos_theta);

		pdf = (cos_theta*M_1_PI);
		
		v = Vector3(cos(phi)*sin_theta, cos_theta, sin(phi)*sin_theta);
	}

	static void cone(Real cos_angle, Vector2 &v, Real &pdf)
	{
	}
	static void cone(Real cos_angle, Vector3 &v, Real &pdf)
	{
		Real cos_thita = 1-RNG::get_real()*(1-cos_angle);
		Real phi = 2*M_PI*RNG::get_real();
		
		Real sin_thita = sqrtf(1-cos_thita*cos_thita);
		v = Vector3(cosf(phi)*sin_theta, cos_theta, sinf(phi)*sin_theta);
		pdf = 1/(2*M_PI*(1-cos_angle));
	}

}; // Sampling


class Sampling :public TSampling<RNG::StdRandom>{};

#endif // _SAMPLING_H_